"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import click
from syndicate.core.build.bundle_processor import load_deploy_output
from syndicate.core.helper import verbose_option

TAGS_GROUP_NAME = 'tags'


@click.group(name=TAGS_GROUP_NAME)
@click.pass_context
def tags(ctx):
    """Manage resources tags"""
    from syndicate.core import PROJECT_STATE, RESOURCES_PROVIDER
    if not PROJECT_STATE.latest_deploy:
        click.echo('No latest deploy')
        raise click.Abort
    deploy_name = PROJECT_STATE.latest_deploy.get('deploy_name')
    bundle_name = PROJECT_STATE.latest_deploy.get('bundle_name')
    output = load_deploy_output(bundle_name, deploy_name)
    ctx.ensure_object(dict)
    ctx.obj['output'] = output
    ctx.obj['tags_resource'] = RESOURCES_PROVIDER.tags_api()


@tags.command(name='apply')
@verbose_option
@click.pass_context
def apply(ctx):
    """Assign tags from config to deployed resources"""
    output = ctx.obj['output']
    ctx.obj['tags_resource'].apply_tags(output)


@tags.command(name='remove')
@verbose_option
@click.pass_context
def remove(ctx):
    """Remove tags from config from deployed resources"""
    output = ctx.obj['output']
    ctx.obj['tags_resource'].remove_tags(output)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.core.resources.dynamo_db_resource import DynamoDBResource, \
    DYNAMODB_TABLE_REQUIRED_PARAMS, AUTOSCALING_REQUIRED_PARAMS
from syndicate.core.resources.helper import validate_params
from syndicate.core.transform.terraform.converter.tf_resource_converter import \
    TerraformResourceConverter
from syndicate.core.transform.terraform.tf_resource_name_builder import \
    build_terraform_resource_name
from syndicate.core.transform.terraform.tf_resource_reference_builder import \
    build_role_arn_ref, build_aws_appautoscaling_target_resource_id_ref, \
    build_aws_appautoscaling_target_scalable_dimension_ref, \
    build_aws_appautoscaling_target_service_namespace_ref


class DynamoDbConverter(TerraformResourceConverter):

    def _add_index_keys_to_definition(self, attributes, index):
        self._append_attr_definition(attributes, index["index_key_name"],
                                     index["index_key_type"])
        if index.get('index_sort_key_name'):
            self._append_attr_definition(attributes,
                                         index["index_sort_key_name"],
                                         index["index_sort_key_type"])

    @staticmethod
    def _append_attr_definition(attributes, attr_name, attr_type):
        for each in attributes:
            if each['name'] == attr_name:
                return
        attributes.append(dict(name=attr_name,
                               type=attr_type))

    def convert(self, name, resource):
        validate_params(name, resource, DYNAMODB_TABLE_REQUIRED_PARAMS)

        hash_key_name = resource.get('hash_key_name')
        hash_key_type = resource.get('hash_key_type')
        sort_key_name = resource.get('sort_key_name')
        sort_key_type = resource.get('sort_key_type')
        read_capacity = resource.get('read_capacity', 1)
        write_capacity = resource.get('write_capacity', 1)
        stream_view_type = resource.get('stream_view_type')

        global_indexes = resource.get('global_indexes', [])
        local_indexes = resource.get('local_indexes', [])

        attributes = self._extract_attributes(global_indexes=global_indexes,
                                              local_indexes=local_indexes,
                                              hash_key_type=hash_key_type,
                                              hash_key=hash_key_name,
                                              sort_key_name=sort_key_name,
                                              sort_key_type=sort_key_type)

        dynamo_db_template = generate_tf_template_for_dynamo_table(
            hash_key=hash_key_name,
            range_key=sort_key_name,
            read_capacity=read_capacity,
            write_capacity=write_capacity,
            table_name=name,
            stream_view_type=stream_view_type,
            attributes=attributes,
            global_indexes=global_indexes,
            local_indexes=local_indexes)
        self.template.add_aws_dynamodb_table(meta=dynamo_db_template)

        autoscaling = resource.get('autoscaling', [])
        for aut in autoscaling:
            validate_params(name, aut, AUTOSCALING_REQUIRED_PARAMS)

            max_capacity = str(aut.get('max_capacity'))
            min_capacity = str(aut.get('min_capacity'))
            dimension = aut.get('dimension')
            resource_name = aut.get('resource_name')
            role_name = aut.get('role_name')

            tf_target_resource_name = build_terraform_resource_name(name,
                                                                    'scalable_target')
            resource_id = DynamoDBResource.build_res_id(dimension=dimension,
                                                        resource_name=resource_name,
                                                        table_name=name)
            target = dynamodb_scalable_target(resource_id=resource_id,
                                              scalable_dimension=dimension,
                                              min_capacity=min_capacity,
                                              max_capacity=max_capacity,
                                              role_name=role_name,
                                              tf_target_resource_name=tf_target_resource_name)

            autoscaling_policy = aut.get('config')
            if autoscaling_policy:
                target_utilization = autoscaling_policy['target_utilization']
                scale_in_cooldown = autoscaling_policy.get('scale_in_cooldown')
                scale_out_cooldown = autoscaling_policy.get(
                    'scale_out_cooldown')
                metric_type = 'DynamoDBWriteCapacityUtilization' \
                    if 'Write' in dimension \
                    else 'DynamoDBReadCapacityUtilization'
                tf_resource_name = build_terraform_resource_name(
                    'dynamodb-test-table_read_policy', resource_name)
                target_policy = dynamo_db_autoscaling_target_policy(
                    target_name=tf_target_resource_name,
                    target_value=target_utilization,
                    predefined_metric_type=metric_type,
                    scale_in_cooldown=scale_in_cooldown,
                    scale_out_cooldown=scale_out_cooldown,
                    tf_resource_name=tf_resource_name)
                self.template.add_aws_appautoscaling_target(meta=target)
                self.template.add_aws_appautoscaling_policy(meta=target_policy)

    def _extract_attributes(self, hash_key,
                            hash_key_type, sort_key_name=None,
                            sort_key_type=None,
                            global_indexes=None, local_indexes=None):
        attributes = [{'name': hash_key,
                       'type': hash_key_type}]

        if sort_key_name:
            attributes.append({'name': sort_key_name, 'type': sort_key_type})

        if global_indexes:
            for index in global_indexes:
                self._add_index_keys_to_definition(attributes=attributes,
                                                   index=index)
        if local_indexes:
            for index in local_indexes:
                self._add_index_keys_to_definition(attributes=attributes,
                                                   index=index)
        return attributes


def generate_tf_template_for_dynamo_table(table_name, hash_key,
                                          range_key,
                                          read_capacity,
                                          write_capacity,
                                          attributes,
                                          stream_view_type=None,
                                          global_indexes=None,
                                          local_indexes=None):
    gl_index_definitions = []
    for gind in global_indexes:
        index = {
            'name': gind.get('name'),
            'hash_key': gind.get('index_key_name'),
            'range_key': gind.get('index_sort_key_name'),
            'projection_type': gind.get('projection_type', 'ALL'),
            'write_capacity': gind.get('write_capacity', 1),
            'read_capacity': gind.get('read_capacity', 1),
            'non_key_attributes': gind.get('non_key_attributes'),
        }
        gl_index_definitions.append(index)

    lc_index_definitions = []
    for loc_ind in local_indexes:
        index = {
            'name': loc_ind.get('name'),
            'range_key': loc_ind.get('index_sort_key_name'),
            'projection_type': loc_ind.get('projection_type', 'ALL'),
            'non_key_attributes': loc_ind.get('non_key_attributes')
        }
        lc_index_definitions.append(index)

    table_content = {
        "name": table_name,
        "hash_key": hash_key,
        "range_key": range_key,
        "read_capacity": read_capacity,
        "write_capacity": write_capacity,
        "global_secondary_index": gl_index_definitions,
        "local_secondary_index": lc_index_definitions,
        "attribute": attributes
    }

    if stream_view_type:
        table_content.update({'stream_enabled': 'true'})
        table_content.update({'stream_view_type': stream_view_type})

    resource = {
        table_name: table_content
    }
    return resource


def dynamodb_scalable_target(tf_target_resource_name, resource_id,
                             scalable_dimension,
                             min_capacity=None,
                             max_capacity=None,
                             role_name=None):
    target = {
        "scalable_dimension": scalable_dimension,
        "service_namespace": 'dynamodb',
        'resource_id': resource_id
    }

    if min_capacity:
        target['max_capacity'] = int(min_capacity)
    if max_capacity:
        target['min_capacity'] = int(max_capacity)
    if role_name:
        target['role_arn'] = build_role_arn_ref(role_name=role_name)

    resource = {
        tf_target_resource_name: target
    }
    return resource


def dynamo_db_autoscaling_target_policy(target_value,
                                        target_name,
                                        tf_resource_name,
                                        predefined_metric_type=None,
                                        resource_label=None,
                                        metric_name=None, namespace=None,
                                        dimensions=None, statistic=None,
                                        unit=None, scale_out_cooldown=None,
                                        scale_in_cooldown=None):
    resource_id = build_aws_appautoscaling_target_resource_id_ref(
        target_name=target_name)
    scalable_dimension = build_aws_appautoscaling_target_scalable_dimension_ref(
        target_name=target_name)
    service_namespace = build_aws_appautoscaling_target_service_namespace_ref(
        target_name=target_name)

    resource = {
        'name': f'dynamodb-read-capacity-utilization-{resource_id}',
        'policy_type': 'TargetTrackingScaling',
        'resource_id': resource_id,
        'scalable_dimension': scalable_dimension,
        'service_namespace': service_namespace,
    }

    target_scaling_config_dict = dict()
    if target_value:
        target_scaling_config_dict['target_value'] = target_value
    predefined_config_dict = dict()
    if predefined_metric_type:
        predefined_config_dict[
            'predefined_metric_type'] = predefined_metric_type
    if resource_label:
        predefined_config_dict['resource_label'] = resource_label
    if predefined_config_dict:
        target_scaling_config_dict[
            'predefined_metric_specification'] = predefined_config_dict

    customized_config_dict = dict()
    if metric_name:
        customized_config_dict['metric_name'] = metric_name
    if namespace:
        customized_config_dict['namespace'] = namespace
    if dimensions:
        customized_config_dict['dimensions'] = dimensions
    if statistic:
        customized_config_dict['statistic'] = statistic
    if unit:
        customized_config_dict['unit'] = unit
    if customized_config_dict:
        target_scaling_config_dict[
            'customized_metric_specification'] = customized_config_dict

    if scale_out_cooldown:
        target_scaling_config_dict['scale_out_cooldown'] = scale_out_cooldown
    if scale_in_cooldown:
        target_scaling_config_dict['scale_in_cooldown'] = scale_in_cooldown
    if target_scaling_config_dict:
        resource[
            'target_tracking_scaling_policy_configuration'] = target_scaling_config_dict

    resource = {
        tf_resource_name: resource
    }
    return resource



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import CLOUD_WATCH_ALARM_TYPE
from syndicate.commons.log_helper import get_logger, get_user_logger

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.cloudwatch_alarm_generator')
USER_LOG = get_user_logger()


class CloudWatchAlarmGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = CLOUD_WATCH_ALARM_TYPE
    CONFIGURATION = {
        "metric_name": None,
        "namespace": None,
        "description": None,
        "period": 1200,  # seconds
        "evaluation_periods": 1,
        "threshold": 1.0,
        "comparison_operator": "GreaterThanOrEqualToThreshold",
        "statistic": "SampleCount",
        "sns_topics": list,
        "lambdas": list,
        "ssm_response_plan": list,
        "evaluate_low_sample_count_percentile": None,
        "datapoints": None
    }

    def _generate_resource_configuration(self) -> dict:
        self._validate_period()
        return super()._generate_resource_configuration()

    def _validate_period(self):
        _LOG.info("Validating given period...")
        to_validate = self._dict.get('period')
        if not to_validate:
            _LOG.warning("The period wasn't given. Skipping validation...")
            return
        if any(to_validate == valid for valid in [10, 30]) \
                or to_validate % 60 == 0:
            _LOG.info(f"Period: '{to_validate}' is valid")
            return
        message = "Valid values for period are 10, 30, or any " \
                  "multiple of 60"
        _LOG.error(f'Period validation error: {message}')
        raise ValueError(message)



"""
    Copyright 2020 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import datetime
import os
import re
from syndicate.core.conf.bucket_view import NAMED_S3_URI_PATTERN
from syndicate.commons.log_helper import get_user_logger

MIN_BUCKET_NAME_LEN = 3
MAX_BUCKET_NAME_LEN = 63

ALL_REGIONS = ['us-east-1', 'us-east-2', 'us-west-1', 'us-west-2', 'sa-east-1',
               'ca-central-1', 'eu-west-1', 'eu-central-1', 'eu-west-2',
               'eu-west-3', 'ap-northeast-1', 'ap-northeast-2', 'ap-east-1',
               'ap-southeast-1', 'ap-southeast-2', 'ap-south-1', 'eu-north-1',
               'eu-south-1', 'ap-northeast-3', 'ap-southeast-3', 'af-south-1']

REQUIRED = 'required'
VALIDATOR = 'validator'

PROJECT_PATH_CFG = 'project_path'
ACCOUNT_ID_CFG = 'account_id'
REGION_CFG = 'region'
LAMBDAS_ALIASES_NAME_CFG = 'lambdas_alias_name'
LOGS_EXPIRATION = 'logs_expiration'

AWS_ACCESS_KEY_ID_CFG = 'aws_access_key_id'
AWS_SECRET_ACCESS_KEY_CFG = 'aws_secret_access_key'
AWS_SESSION_TOKEN_CFG = 'aws_session_token'
DEPLOY_TARGET_BUCKET_CFG = 'deploy_target_bucket'
PROJECTS_MAPPING_CFG = 'build_projects_mapping'
RESOURCES_SUFFIX_CFG = 'resources_suffix'
RESOURCES_PREFIX_CFG = 'resources_prefix'
IAM_SUFFIX_CFG = 'iam_suffix'
EXTENDED_PREFIX_MODE_CFG = 'extended_prefix_mode'
EXTENDED_PREFIX_PATTERN = '^[a-z0-9-]+$'

USE_TEMP_CREDS_CFG = 'use_temp_creds'
SERIAL_NUMBER_CFG = 'serial_number'
TEMP_AWS_ACCESS_KEY_ID_CFG = 'temp_aws_access_key_id'
TEMP_AWS_SECRET_ACCESS_KEY_CFG = 'temp_aws_secret_access_key'
TEMP_AWS_SESSION_TOKEN_CFG = 'temp_aws_session_token'
EXPIRATION_CFG = 'expiration'
SESSION_DURATION_CFG = 'session_duration'
ACCESS_ROLE_CFG = 'access_role'
IAM_PERMISSIONS_BOUNDARY_CFG = 'iam_permissions_boundary'
LOCK_LIFETIME_MINUTES_CFG = 'lock_lifetime_minutes'

TAGS_CFG = 'tags'

PYTHON_LANGUAGE_NAME = 'python'
NODEJS_LANGUAGE_NAME = 'nodejs'
JAVA_LANGUAGE_NAME = 'java'
SWAGGER_UI_NAME = 'swagger_ui'

ALLOWED_RUNTIME_LANGUAGES = [PYTHON_LANGUAGE_NAME,
                             JAVA_LANGUAGE_NAME,
                             NODEJS_LANGUAGE_NAME,
                             SWAGGER_UI_NAME]

REQUIRED_PARAM_ERROR = 'The required key {} is missing'
UNKNOWN_PARAM_MESSAGE = 'Unknown parameter(s) in the configuration file: {}'

USER_LOG = get_user_logger()


class ConfigValidator:

    def __init__(self, config_dict) -> None:
        self._config_dict = config_dict
        self._extended_prefix_mode = config_dict.get(EXTENDED_PREFIX_MODE_CFG)
        self._fields_validators_mapping = {
            PROJECT_PATH_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_project_path},
            ACCOUNT_ID_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_account_id},
            REGION_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_region},
            DEPLOY_TARGET_BUCKET_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_bundle_bucket_name},
            PROJECTS_MAPPING_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_project_mapping},
            AWS_ACCESS_KEY_ID_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_access_key},
            AWS_SECRET_ACCESS_KEY_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_secret_access_key},
            AWS_SESSION_TOKEN_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_session_token},
            RESOURCES_PREFIX_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_resource_prefix},
            RESOURCES_SUFFIX_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_resources_prefix_suffix},
            IAM_SUFFIX_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_resources_prefix_suffix
            },
            EXTENDED_PREFIX_MODE_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_extended_prefix_mode
            },
            USE_TEMP_CREDS_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_use_temp_creds
            },
            SERIAL_NUMBER_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_serial_number
            },
            SESSION_DURATION_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_session_duration,
            },
            TEMP_AWS_SECRET_ACCESS_KEY_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_access_key
            },
            TEMP_AWS_ACCESS_KEY_ID_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_secret_access_key
            },
            TEMP_AWS_SESSION_TOKEN_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_session_token
            },
            EXPIRATION_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_expiration
            },
            ACCESS_ROLE_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_access_role
            },
            TAGS_CFG: {
                REQUIRED: False,
                VALIDATOR: self.validate_tags
            },
            IAM_PERMISSIONS_BOUNDARY_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_iam_permissions_boundary
            },
            LOCK_LIFETIME_MINUTES_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_lock_lifetime_minutes
            }
        }

    def validate(self):
        error_messages = {}
        unknown_params = set(self._config_dict.keys()) - set(
            self._fields_validators_mapping.keys())
        if unknown_params:
            USER_LOG.warn(UNKNOWN_PARAM_MESSAGE.format(unknown_params))

        for key, validation_rules in self._fields_validators_mapping.items():
            value = self._config_dict.get(key)
            is_required = validation_rules.get(REQUIRED)
            if is_required and not value:
                error_messages[key] = REQUIRED_PARAM_ERROR.format(key)
                continue
            if value:
                validator_func = validation_rules.get(VALIDATOR)
                validation_errors = validator_func(key, value)
                if validation_errors:
                    error_messages[key] = validation_errors
        return error_messages

    def _validate_project_path(self, key, value):
        str_error = self._assert_value_is_str(key, value)
        if str_error:
            return [str_error]
        errors = []
        if len(value) == 0:
            errors.append(f'{key} must not be empty')
        if not os.path.exists(value):
            errors.append(f'The path {value} specified in {key} must exist')
        return errors

    @staticmethod
    def _validate_account_id(key, value):
        errors = []
        try:
            int(value)
        except TypeError:
            errors.append(f'{key} must be int, not {type(value)}')
            return errors
        if len(str(value)) != 12:
            errors.append(f'{key} must be a 12-digit number')
        return errors

    def _validate_region(self, key, value):
        str_error = self._assert_value_is_str(key, value)
        if str_error:
            return [str_error]
        if value not in ALL_REGIONS:
            return [
                f'{key} value must be one of {ALL_REGIONS}, but is {value}'
            ]

    def _validate_bundle_bucket_name(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]

        errors = []
        # check min length
        name = re.compile(
            NAMED_S3_URI_PATTERN).match(value).groupdict().get('name')
        if len(name) < MIN_BUCKET_NAME_LEN or len(name) > MAX_BUCKET_NAME_LEN:
            errors.append(f'The length of {key} must be between '
                          f'{MIN_BUCKET_NAME_LEN} and {MAX_BUCKET_NAME_LEN} '
                          f'characters long but not {len(name)}')
        return errors

    def _validate_project_mapping(self, key, value):
        errors = []
        if type(value) is not dict:
            errors.append(f'{key} must be type of dict')
            return errors
        project_path = self._config_dict.get(PROJECT_PATH_CFG)
        for key in value.keys():
            if key not in ALLOWED_RUNTIME_LANGUAGES:
                errors.append(f'{key} is not supported to be built')
                continue
            for build_key, paths in value.items():
                if not paths:
                    errors.append(f'The path in {build_key} project '
                                  f'mapping not specified')

                else:
                    for path in paths:
                        if not os.path.exists(os.path.join(
                                project_path, path)):
                            errors.append(
                                f'The path in {key}:{build_key} project '
                                f'mapping does not exists: {path}')
        return errors

    def _validate_aws_access_key(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]
        if len(value) < 16 or len(value) > 128:
            return [
                f'The length of {key} must be in a '
                f'range between 16 and 128 characters']

    def _validate_aws_secret_access_key(self, key, value):
        # the only constraint found
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]

    def _validate_aws_session_token(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]

    def _validate_use_temp_creds(self, key, value):
        bool_error = self._assert_value_is_bool(
            key=key, value=value
        )
        if bool_error:
            return [bool_error]

    def _validate_serial_number(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]

    def _validate_access_role(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]

    @staticmethod
    def validate_tags(key, value):
        errors = []
        if not value:
            return errors
        if not isinstance(value, dict):
            errors.append(f'\'{key}\' param must be a dictionary but '
                          f'not a \'{type(value).__name__}\'')
            return errors
        if len(value) > 50:
            errors.append(f'Each resource can have up to 50 user created tags.'
                          f' You have specified: {len(value)}')
        for tag_name, tag_value in value.items():
            if tag_name.startswith('aws:'):
                errors.append(f'\'{tag_name}\': you can\'t create, edit or '
                              f'delete a tag that begins with the \'aws:\' '
                              f'prefix.')
                if not 1 <= len(tag_name) <= 128:
                    errors.append(f'\'{tag_name}\': the tag key must be a '
                                  f'minimum of 1 and a maximum of 128 Unicode '
                                  f'characters')
                if not 0 <= len(tag_value) <= 256:
                    errors.append(f'\'{tag_value}\': the tag value must be a '
                                  f'minimum of 0 and a maximum of 256 Unicode '
                                  f'characters')
        return errors

    @staticmethod
    def _validate_iam_permissions_boundary(key, value):
        errors = []
        if not isinstance(value, str):
            return [f'\'{key}\' must have a string type']
        return errors

    @staticmethod
    def _validate_expiration(key, value):
        if not isinstance(value, datetime.datetime):
            return [f'\'{key}\' must be a valid ISO 8601 format string']
        return []

    @staticmethod
    def _validate_resources_prefix_suffix(key, value):
        str_error = ConfigValidator._assert_value_is_str(key=key, value=value)
        if str_error:
            return [str_error]
        if len(value) > 5:
            return [
                f'The length of {key} must be less or equal to 5 character']

    def _validate_resource_prefix(self, key, value):
        if self._extended_prefix_mode:
            return self._validate_resources_prefix_extended_mode(key, value)
        return self._validate_resources_prefix_suffix(key, value)

    @staticmethod
    def _validate_extended_prefix_mode(key, value):
        bool_error = ConfigValidator._assert_value_is_bool(
            key=key, value=value
        )
        if bool_error:
            return [bool_error]

    @staticmethod
    def _validate_resources_prefix_extended_mode(key, value):
        result = []
        str_error = ConfigValidator._assert_value_is_str(key=key, value=value)
        if str_error:
            result.append(str_error)
        if len(value) > 14:
            result.append(f'The length of {key} must be less or equal to 14 '
                          f'character')
        if '--' in value:
            result.append(f'The {key} must not contain two consecutive '
                          f'hyphens')
        if not value[0].isalpha():
            result.append(f'The first character of the {key} must be a letter')
        if not re.match(EXTENDED_PREFIX_PATTERN, value):
            result.append(f'The {key} must contain only lowercase letters, '
                          f'numbers, and hyphens')
        return result

    @staticmethod
    def validate_prefix_suffix(key, value):
        result = ConfigValidator._validate_resources_prefix_suffix(key, value)
        if result:
            return result[0]

    @staticmethod
    def validate_extended_prefix(key, value):
        result = ConfigValidator._validate_resources_prefix_extended_mode(
            key, value)
        if result:
            return result

    @staticmethod
    def _validate_session_duration(key, value):
        if not isinstance(value, int):
            return [f'\'{key}\' must a an integer']
        if value < 900:
            return [f'\'{key}\' must begin from 900 seconds']

    @staticmethod
    def _validate_lock_lifetime_minutes(key, value):
        if not isinstance(value, int):
            return [f'\'{key}\' must a an integer']
        if not 0 <= value <= 300:
            return [f'\'{key}\' value must be between 0 and 300 minutes']

    @staticmethod
    def _assert_value_is_str(key, value):
        if type(value) is not str:
            return f'{key} must be type of string'

    @staticmethod
    def _assert_value_is_bool(key, value):
        if type(value) is not bool:
            return f'{key} must be type of bool'



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import awslambda

from syndicate.core.constants import S3_PATH_NAME
from syndicate.core.resources.helper import validate_params
from syndicate.core.resources.lambda_resource import \
    LAMBDA_LAYER_REQUIRED_PARAMS
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import lambda_layer_logic_name


class CfLambdaLayerConverter(CfResourceConverter):

    def convert(self, name, meta):
        validate_params(name, meta, LAMBDA_LAYER_REQUIRED_PARAMS)

        logic_name = lambda_layer_logic_name(name)
        layer = awslambda.LayerVersion(logic_name)

        layer.Content = awslambda.Content(
            S3Bucket=self.config.deploy_target_bucket,
            S3Key=meta[S3_PATH_NAME])
        layer.LayerName = name
        layer.CompatibleRuntimes = meta['runtimes']

        if meta.get('description'):
            layer.Description = meta['description']
        if meta.get('license'):
            layer.LicenseInfo = meta['license']
        self.template.add_resource(layer)



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
import click
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import IAM_ROLE, IAM_POLICY
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.iam_role_generator')
USER_LOG = get_user_logger()


class IAMRoleGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = IAM_ROLE
    CONFIGURATION = {
        "principal_service": None,
        "predefined_policies": list,
        "custom_policies": list,
        "allowed_accounts": list,
        "external_id": None,
        "instance_profile": bool,
        "permissions_boundary": str
    }

    def _resolve_configuration(self, defaults_dict=None) -> dict:
        self.validate_custom_policies_existence()

        if self._dict['principal_service'] == 'ec2' and \
                not self._dict['instance_profile']:
            USER_LOG.info("Setting 'instance_profile' value to 'True' because "
                          "it wasn't specified and principal service is 'ec2'")
            self._dict['instance_profile'] = True
        return super()._resolve_configuration()

    def validate_custom_policies_existence(self):
        custom_policies = set(self._dict.get('custom_policies', []))
        _LOG.info(f"Validating existence of these policies: {custom_policies}")
        available_policies_dict = self._find_resources_by_type(IAM_POLICY)

        available_policies = set()
        for value in available_policies_dict.values():
            available_policies.update(value)

        custom_policies = custom_policies - available_policies
        if custom_policies:
            message = f"Custom policies: {custom_policies} was not found " \
                      f"in deployment resources"
            _LOG.error(f"Validation error: {message}")
            raise ValueError(message)
        _LOG.info(f"Validation successfully finished, policies exist")



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import cloudwatch

from syndicate.core.resources.cloud_watch_alarm_resource import \
    CLOUDWATCH_ALARM_REQUIRED_PARAMS
from syndicate.core.resources.helper import validate_params
from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import to_logic_name, sns_topic_logic_name


class CfCloudWatchAlarmConverter(CfResourceConverter):

    def convert(self, name, meta):
        validate_params(name, meta, CLOUDWATCH_ALARM_REQUIRED_PARAMS)

        alarm = cloudwatch.Alarm(to_logic_name('CloudWatchAlarm', name))

        alarm.AlarmName = name
        alarm.ComparisonOperator = meta['comparison_operator']
        alarm.EvaluationPeriods = meta['evaluation_periods']
        alarm.MetricName = meta['metric_name']
        alarm.Namespace = meta['namespace']
        alarm.Period = meta['period']
        alarm.Statistic = meta['statistic']
        alarm.Threshold = meta['threshold']

        sns_topics = meta.get('sns_topics')
        if sns_topics:
            sns_topic_arns = []
            for each in sns_topics:
                topic = self.get_resource(sns_topic_logic_name(each))
                sns_topic_arns.append(topic.ref())
                alarm.AlarmActions = sns_topic_arns

        self.template.add_resource(alarm)



from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import SNS_PLATFORM_APPLICATION_TYPE


class SNSApplicationGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = SNS_PLATFORM_APPLICATION_TYPE
    CONFIGURATION = {
        "platform": None,
        "region": None,
        "attributes": dict
    }

    def __init__(self, **kwargs):
        self.attributes = []
        if 'attributes' in kwargs:
            self.attributes.extend(kwargs.pop('attributes'))
        super().__init__(**kwargs)

    def _resolve_configuration(self, defaults_dict=None) -> dict:
        result = super()._resolve_configuration()
        for attr_name, attr_value in self.attributes:
            result['attributes'][attr_name] = attr_value
        return result



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.commons.log_helper import get_logger
from syndicate.core.build.validator.batch_compenv_validator import validate_batch_compenv
from syndicate.core.build.validator.dynamodb_validator import \
    validate_dynamodb, validate_dax_cluster
from syndicate.core.build.validator.batch_jobdef_validator import validate_batch_jobdef
from syndicate.core.build.validator.lambda_validator import validate_lambda
from syndicate.core.build.validator.ec2_launch_template_validator import \
    validate_launch_template
from syndicate.core.constants import \
    (LAMBDA_CONFIG_FILE_NAME, LAMBDA_TYPE, DYNAMO_TABLE_TYPE,
     BATCH_COMPENV_TYPE, BATCH_JOBDEF_TYPE, DAX_CLUSTER_TYPE,
     EC2_LAUNCH_TEMPLATE_TYPE, RESOURCE_LIST)

ALL_TYPES = 'all_types'

_LOG = get_logger('validator')


def common_validate(resource_name, resource_meta, all_meta):
    dependencies = resource_meta.get('dependencies')
    if dependencies:
        for dependency in dependencies:
            errors = []

            if not dependency.get('resource_name'):
                errors.append(
                    f"There is no 'resource_name' in resource "
                    f"'{resource_name}' dependency {dependency}")
            elif dependency.get('resource_name') not in list(all_meta.keys()):
                errors.append(
                    f"The resource '{resource_name}' depends on resource "
                    f"'{dependency.get('resource_name')}' that is not a part "
                    f"of the project. Please double-check the project "
                    f"resources description add it to the project or remove "
                    f"it from the resource '{resource_name}' dependencies.")

            if not dependency.get('resource_type'):
                errors.append(
                    f"There is no 'resource_type' in resource "
                    f"'{resource_name}' dependency {dependency}")
            elif dependency.get('resource_type') not in RESOURCE_LIST:
                errors.append(
                    f"Unsupported resource type "
                    f"'{dependency.get('resource_type')}' found in "
                    f"the resource '{resource_name}' dependency "
                    f"'{dependency}'.")
            if errors:
                raise AssertionError(str(errors))


# validation customization
VALIDATOR_BY_TYPE_MAPPING = {
    ALL_TYPES: common_validate,
    DYNAMO_TABLE_TYPE: validate_dynamodb,
    DAX_CLUSTER_TYPE: validate_dax_cluster,
    BATCH_COMPENV_TYPE: validate_batch_compenv,
    BATCH_JOBDEF_TYPE: validate_batch_jobdef,
    LAMBDA_TYPE: validate_lambda,
    EC2_LAUNCH_TEMPLATE_TYPE: validate_launch_template
}



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.core.build.artifact_processor import RUNTIME_NODEJS
from syndicate.core.conf.validator import (
    LAMBDAS_ALIASES_NAME_CFG, LOGS_EXPIRATION
)
from syndicate.core.generators import (_alias_variable,
                                       FILE_LAMBDA_HANDLER_NODEJS)
from syndicate.core.groups import DEFAULT_RUNTIME_VERSION

POLICY_LAMBDA_BASIC_EXECUTION = "lambda-basic-execution"

LAMBDA_ROLE_NAME_PATTERN = '{0}-role'  # 0 - lambda_name

SRC_MAIN_JAVA = 'jsrc/main/java'
FILE_POM = 'pom.xml'
CANCEL_MESSAGE = 'Creating of {} has been canceled.'

JAVA_LAMBDA_HANDLER_CLASS = """package {java_package_name};

import com.amazonaws.services.lambda.runtime.Context;
import com.amazonaws.services.lambda.runtime.RequestHandler;
import com.syndicate.deployment.annotations.lambda.LambdaHandler;
import com.syndicate.deployment.model.RetentionSetting;

import java.util.HashMap;
import java.util.Map;

@LambdaHandler(
    lambdaName = "{lambda_name}",
	roleName = "{lambda_role_name}",
	isPublishVersion = true,
	aliasName = "${lambdas_alias_name}",
	logsExpiration = RetentionSetting.SYNDICATE_ALIASES_SPECIFIED
)
public class {lambda_class_name} implements RequestHandler<Object, Map<String, Object>> {

	public Map<String, Object> handleRequest(Object request, Context context) {
		System.out.println("Hello from lambda");
		Map<String, Object> resultMap = new HashMap<String, Object>();
		resultMap.put("statusCode", 200);
		resultMap.put("body", "Hello from Lambda");
		return resultMap;
	}
}
"""

JAVA_ROOT_POM_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0"
         xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
         xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>

    <groupId>{project_name}-group</groupId>
    <artifactId>{project_name}</artifactId>
    <version>1.0.0</version>

    <properties>
        <maven-shade-plugin.version>3.5.2</maven-shade-plugin.version>
        <syndicate.java.plugin.version>1.13.0</syndicate.java.plugin.version>
        <maven.compiler.source>11</maven.compiler.source>
        <maven.compiler.target>11</maven.compiler.target>
        <project.build.sourceEncoding>UTF-8</project.build.sourceEncoding>
        <src.dir>jsrc/main/java</src.dir>
        <resources.dir>jsrc/main/resources</resources.dir>
    </properties>

    <dependencies>
        <!-- AWS dependencies-->
        <dependency>
            <groupId>com.amazonaws</groupId>
            <artifactId>aws-lambda-java-core</artifactId>
            <version>1.2.0</version>
        </dependency>
        <!--Syndicate annotations-->
        <dependency>
            <groupId>net.sf.aws-syndicate</groupId>
            <artifactId>deployment-configuration-annotations</artifactId>
            <version>${syndicate.java.plugin.version}</version>
        </dependency>
    </dependencies>

    <build>
        <sourceDirectory>${src.dir}</sourceDirectory>
        <resources>
            <resource>
                <directory>${resources.dir}</directory>
            </resource>
        </resources>        
        <plugins>
            <plugin>
                <groupId>net.sf.aws-syndicate</groupId>
                <artifactId>deployment-configuration-maven-plugin</artifactId>
                <version>${syndicate.java.plugin.version}</version>
                <configuration>
                    <packages>
                        <!--packages to scan-->
                    </packages>
                    <fileName>${project.name}-${project.version}.jar</fileName>
                </configuration>
                <executions>
                    <execution>
                        <id>generate-config</id>
                        <phase>compile</phase>
                        <inherited>false</inherited>
                        <goals>
                            <goal>gen-deployment-config</goal>
                            <goal>assemble-lambda-layer-files</goal>
                        </goals>
                    </execution>
                </executions>
            </plugin>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-shade-plugin</artifactId>
                <version>${maven-shade-plugin.version}</version>
                <configuration>
                    <filters>
                        <filter>
                            <artifact>*:*</artifact>
                            <excludes>
                                <exclude>META-INF/*.SF</exclude>
                                <exclude>META-INF/*.DSA</exclude>
                                <exclude>META-INF/*.RSA</exclude>
                            </excludes>
                        </filter>
                    </filters>
                    <createDependencyReducedPom>false</createDependencyReducedPom>
                </configuration>
                <executions>
                    <execution>
                        <phase>package</phase>
                        <goals>
                            <goal>shade</goal>
                        </goals>
                    </execution>
                </executions>
            </plugin>
        </plugins>
    </build>

</project>
"""

PYTHON_LAMBDA_HANDLER_TEMPLATE = """from commons.log_helper import get_logger
from commons.abstract_lambda import AbstractLambda

_LOG = get_logger('LambdaName-handler')


class LambdaName(AbstractLambda):

    def validate_request(self, event) -> dict:
        pass
        
    def handle_request(self, event, context):
        \"\"\"
        Explain incoming event here
        \"\"\"
        # todo implement business logic
        return 200
    

HANDLER = LambdaName()


def lambda_handler(event, context):
    return HANDLER.lambda_handler(event=event, context=context)
"""

NODEJS_LAMBDA_HANDLER_TEMPLATE = """exports.handler = async (event) => {
    // TODO implement
    const response = {
        statusCode: 200,
        body: JSON.stringify('Hello from Lambda!'),
    };
    return response;
};
"""

GITIGNORE_CONTENT = """.syndicate
logs/
.syndicate-config-*/
"""

CHANGELOG_TEMPLATE = """# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),
and this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).

## [1.0.0] - yyyy-MM-dd
### Added
    -  Added items

### Changed
    -  Changed items 

### Removed
    -  Removed items 
"""

README_TEMPLATE = """# project_name

High level project overview - business value it brings, non-detailed technical overview.

### Notice
All the technical details described below are actual for the particular
version, or a range of versions of the software.
### Actual for versions: 1.0.0

## project_name diagram

![project_name](pics/project_name_diagram.png)

## Lambdas descriptions

### Lambda `lambda-name`
Lambda feature overview.

### Required configuration
#### Environment variables
* environment_variable_name: description

#### Trigger event
```buildoutcfg
{
    "key": "value",
    "key1": "value1",
    "key2": "value3"
}
```
* key: [Required] description of key
* key1: description of key1

#### Expected response
```buildoutcfg
{
    "status": 200,
    "message": "Operation succeeded"
}
```
---

## Deployment from scratch
1. action 1 to deploy the software
2. action 2
...

"""

ABSTRACT_LAMBDA_CONTENT = """from abc import abstractmethod

from commons import ApplicationException, build_response
from commons.log_helper import get_logger

_LOG = get_logger('abstract-lambda')


class AbstractLambda:

    @abstractmethod
    def validate_request(self, event) -> dict:
        \"\"\"
        Validates event attributes
        :param event: lambda incoming event
        :return: dict with attribute_name in key and error_message in value
        \"\"\"
        pass

    @abstractmethod
    def handle_request(self, event, context):
        \"\"\"
        Inherited lambda function code
        :param event: lambda event
        :param context: lambda context
        :return:
        \"\"\"
        pass

    def lambda_handler(self, event, context):
        try:
            _LOG.debug(f'Request: {event}')
            if event.get('warm_up'):
                return
            errors = self.validate_request(event=event)
            if errors:
                return build_response(code=400,
                                      content=errors)
            execution_result = self.handle_request(event=event,
                                                   context=context)
            _LOG.debug(f'Response: {execution_result}')
            return execution_result
        except ApplicationException as e:
            _LOG.error(f'Error occurred; Event: {event}; Error: {e}')
            return build_response(code=e.code,
                                  content=e.content)
        except Exception as e:
            _LOG.error(
                f'Unexpected error occurred; Event: {event}; Error: {e}')
            return build_response(code=500,
                                  content='Internal server error')
"""

INIT_CONTENT = """from commons.exception import ApplicationException

RESPONSE_BAD_REQUEST_CODE = 400
RESPONSE_UNAUTHORIZED = 401
RESPONSE_FORBIDDEN_CODE = 403
RESPONSE_RESOURCE_NOT_FOUND_CODE = 404
RESPONSE_OK_CODE = 200
RESPONSE_INTERNAL_SERVER_ERROR = 500
RESPONSE_NOT_IMPLEMENTED = 501
RESPONSE_SERVICE_UNAVAILABLE_CODE = 503


def build_response(content, code=200):
    if code == RESPONSE_OK_CODE:
        return {
            'code': code,
            'body': content
        }
    raise ApplicationException(
        code=code,
        content=content
    )


def raise_error_response(code, content):
    raise ApplicationException(code=code, content=content)
"""

EXCEPTION_CONTENT = """class ApplicationException(Exception):

    def __init__(self, code, content):
        self.code = code
        self.content = content

    def __str__(self):
        return f'{self.code}:{self.content}'
"""

LOG_HELPER_CONTENT = """import logging
import os
from sys import stdout

_name_to_level = {
    'CRITICAL': logging.CRITICAL,
    'FATAL': logging.FATAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG
}

logger = logging.getLogger(__name__)
logger.propagate = False
console_handler = logging.StreamHandler(stream=stdout)
console_handler.setFormatter(
    logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s'))
logger.addHandler(console_handler)


log_level = _name_to_level.get(os.environ.get('log_level'))
if not log_level:
    log_level = logging.INFO
logging.captureWarnings(True)


def get_logger(log_name, level=log_level):
    module_logger = logger.getChild(log_name)
    if level:
        module_logger.setLevel(level)
    return module_logger
"""

PYTHON_TESTS_INIT_CONTENT = \
"""import sys
from pathlib import Path

SOURCE_FOLDER = 'src'


class ImportFromSourceContext:
    \"\"\"Context object to import lambdas and packages. It's necessary because
    root path is not the path to the syndicate project but the path where
    lambdas are accumulated - SOURCE_FOLDER \"\"\"

    def __init__(self, source_folder=SOURCE_FOLDER):
        self.source_folder = source_folder
        self.assert_source_path_exists()

    @property
    def project_path(self) -> Path:
        return Path(__file__).parent.parent

    @property
    def source_path(self) -> Path:
        return Path(self.project_path, self.source_folder)

    def assert_source_path_exists(self):
        source_path = self.source_path
        if not source_path.exists():
            print(f'Source path "{source_path}" does not exist.',
                  file=sys.stderr)
            sys.exit(1)

    def _add_source_to_path(self):
        source_path = str(self.source_path)
        if source_path not in sys.path:
            sys.path.append(source_path)

    def _remove_source_from_path(self):
        source_path = str(self.source_path)
        if source_path in sys.path:
            sys.path.remove(source_path)

    def __enter__(self):
        self._add_source_to_path()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._remove_source_from_path()

"""

PYTHON_TESTS_INIT_LAMBDA_TEMPLATE = \
"""import unittest
import importlib
from tests import ImportFromSourceContext

with ImportFromSourceContext():
    LAMBDA_HANDLER = importlib.import_module('lambdas.{lambda_name}.handler')


class {camel_lambda_name}LambdaTestCase(unittest.TestCase):
    \"\"\"Common setups for this lambda\"\"\"

    def setUp(self) -> None:
        self.HANDLER = LAMBDA_HANDLER.{camel_lambda_name}()

"""

PYTHON_TESTS_BASIC_TEST_CASE_TEMPLATE = \
"""from tests.{test_lambda_folder} import {camel_lambda_name}LambdaTestCase


class TestSuccess({camel_lambda_name}LambdaTestCase):

    def test_success(self):
        self.assertEqual(self.HANDLER.handle_request(dict(), dict()), 200)

"""

S3_BUCKET_PUBLIC_READ_POLICY = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Sid": "PublicReadGetObject",
            "Effect": "Allow",
            "Principal": "*",
            "Action": [
                "s3:GetObject"
            ],
            "Resource": [
                "arn:aws:s3:::{bucket_name}/*"
            ]
        }
    ]
}

SWAGGER_UI_INDEX_FILE_CONTENT = \
    """<!DOCTYPE html>
    <html lang="en">
      <head>
        <meta charset="utf-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1" />
        <meta
          name="description"
          content="SwaggerUI"
        />
        <title>SwaggerUI</title>
        <link rel="stylesheet" href="https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui.css" />
      </head>
      <body>
      <div id="swagger-ui"></div>
      <script src="https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui-bundle.js" crossorigin></script>
      <script src="https://unpkg.com/swagger-ui-dist@4.5.0/swagger-ui-standalone-preset.js" crossorigin></script>
      <script>
        window.onload = () => {
          window.ui = SwaggerUIBundle({
            url: './spec_file_name',
            dom_id: '#swagger-ui',
            presets: [
              SwaggerUIBundle.presets.apis,
              SwaggerUIStandalonePreset
            ],
            layout: "StandaloneLayout",
          });
        };
      </script>
      </body>
    </html>"""

REQUIREMENTS_FILE_CONTENT = '# list of requirements'
LOCAL_REQUIREMENTS_FILE_CONTENT = '# local requirements'


def _stringify(dict_content):
    return json.dumps(dict_content, indent=2)


def _generate_python_node_lambda_config(lambda_name, lambda_relative_path):
    return _stringify({
        'version': '1.0',
        'name': lambda_name,
        'func_name': 'handler.lambda_handler',
        'resource_type': 'lambda',
        'iam_role_name': LAMBDA_ROLE_NAME_PATTERN.format(lambda_name),
        'runtime': 'python3.10',
        'memory': 128,
        'timeout': 100,
        'lambda_path': lambda_relative_path,
        'dependencies': [],
        'event_sources': [],
        'env_variables': {},
        'publish_version': True,
        'alias': _alias_variable(LAMBDAS_ALIASES_NAME_CFG),
        'url_config': {},
        'ephemeral_storage': 512,
        'logs_expiration': _alias_variable(LOGS_EXPIRATION)
        # 'platforms': ['manylinux2014_x86_64']
        # by default (especially if you have linux), you don't need it
    })


def _generate_python_node_layer_config(layer_name, runtime):
    return _stringify({
        "name": layer_name,
        "resource_type": "lambda_layer",
        "runtimes": [
            DEFAULT_RUNTIME_VERSION.get(runtime)
        ],
        "deployment_package": f"{layer_name}_layer.zip"
    })


def _generate_node_layer_package_file(layer_name):
    return _stringify({
        "name": layer_name,
        "version": "1.0.0",
        "description": "",
        "main": "index.js",
        "scripts": {},
        "author": "",
        "license": "ISC",
        "dependencies": {}
    })


def _generate_node_layer_package_lock_file(layer_name):
    return _stringify({
            "name": layer_name,
            "version": "1.0.0",
            "lockfileVersion": 1,
            "requires": True,
            "dependencies": {}
        })


def _generate_nodejs_node_lambda_config(lambda_name, lambda_relative_path):
    return _stringify({
        'version': '1.0',
        'name': lambda_name,
        'func_name': f'lambdas/{lambda_name}/index.handler',
        'resource_type': 'lambda',
        'iam_role_name': LAMBDA_ROLE_NAME_PATTERN.format(lambda_name),
        'runtime': RUNTIME_NODEJS,
        'memory': 128,
        'timeout': 100,
        'lambda_path': lambda_relative_path,
        'dependencies': [],
        'event_sources': [],
        'env_variables': {},
        'publish_version': True,
        'alias': _alias_variable(LAMBDAS_ALIASES_NAME_CFG),
        'url_config': {},
        'ephemeral_storage': 512
    })


def _generate_package_nodejs_lambda(lambda_name):
    return _stringify({
        "name": lambda_name,
        "version": "1.0.0",
        "description": "",
        "main": FILE_LAMBDA_HANDLER_NODEJS,
        "scripts": {},
        "author": "",
        "license": "ISC",
        "dependencies": {
        }
    })


def _generate_package_lock_nodejs_lambda(lambda_name):
    return _stringify({
        "name": lambda_name,
        "version": "1.0.0",
        "lockfileVersion": 1,
        "requires": True,
        "dependencies": {}
    })


def _get_lambda_default_policy():
    return _stringify({
        POLICY_LAMBDA_BASIC_EXECUTION: {
            'policy_content': {
                "Statement": [
                    {
                        "Action": [
                            "logs:CreateLogGroup",
                            "logs:CreateLogStream",
                            "logs:PutLogEvents",
                            "dynamodb:GetItem",
                            "dynamodb:Query",
                            "dynamodb:PutItem",
                            "dynamodb:Batch*",
                            "dynamodb:DeleteItem",
                            "ssm:PutParameter",
                            "ssm:GetParameter",
                            "kms:Decrypt"
                        ],
                        "Effect": "Allow",
                        "Resource": "*"
                    }
                ],
                "Version": "2012-10-17"},
            "resource_type": "iam_policy"
        }
    })


def _generate_lambda_role_config(role_name, stringify=True):
    role_content = {
        role_name: {
            "predefined_policies": [],
            "principal_service": "lambda",
            "custom_policies": [
                POLICY_LAMBDA_BASIC_EXECUTION
            ],
            "resource_type": "iam_role"
        }
    }
    return _stringify(role_content) if stringify else role_content


def _generate_swagger_ui_config(resource_name, path_to_spec, target_bucket):
    return _stringify({
            "name": resource_name,
            "resource_type": "swagger_ui",
            "path_to_spec": path_to_spec,
            "target_bucket": target_bucket
        })



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.cognito_identity_provider_connection')


@apply_methods_decorator(retry())
class CognitoIdentityProviderConnection(object):
    """ Cognito identity provider connection class."""

    def __init__(self, client_config, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = client('cognito-idp', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token,
                             config=client_config)
        _LOG.debug('Opened new Cognito identity connection.')

    def create_user_pool(self, pool_name, auto_verified_attributes=None,
                         sms_configuration=None, username_attributes=None,
                         policies=None):
        """
        Crete Cognito user pool and get user pool id.
        """
        params = dict(PoolName=pool_name)
        if auto_verified_attributes:
            params['AutoVerifiedAttributes'] = auto_verified_attributes
        if sms_configuration:
            params['SmsConfiguration'] = sms_configuration
        if username_attributes:
            params['UsernameAttributes'] = username_attributes
        if policies:
            params['Policies'] = policies

        response = self.client.create_user_pool(**params)
        return response['UserPool'].get('Id')

    def create_user_pool_client(
            self, user_pool_id, client_name, generate_secret=True,
            refresh_token_validity=None,
            read_attributes=None,
            write_attributes=None, explicit_auth_flows=None,
            supported_identity_providers=None,
            callback_urls=None, logout_urls=None, default_redirect_uri=None,
            allowed_oauth_flows=None, allowed_oauth_scopes=None,
            allowed_oauth_flows_user_pool_client=None,
            analytics_configuration=None, prevent_user_existence_errors=None,
            enable_token_revocation=None):
        params = dict(UserPoolId=user_pool_id, ClientName=client_name,
                      GenerateSecret=generate_secret)
        if refresh_token_validity:
            params.update(RefreshTokenValidity=refresh_token_validity)
        if read_attributes:
            params.update(ReadAttributes=read_attributes)
        if write_attributes:
            params.update(WriteAttributes=write_attributes)
        if explicit_auth_flows:
            params.update(ExplicitAuthFlows=explicit_auth_flows)
        if supported_identity_providers:
            params.update(
                SupportedIdentityProviders=supported_identity_providers)
        if callback_urls:
            params.update(CallbackURLs=callback_urls)
        if logout_urls:
            params.update(LogoutURLs=logout_urls)
        if default_redirect_uri:
            params.update(DefaultRedirectURI=default_redirect_uri)
        if allowed_oauth_flows:
            params.update(AllowedOAuthFlows=allowed_oauth_flows)
        if allowed_oauth_scopes:
            params.update(AllowedOAuthScopes=allowed_oauth_scopes)
        if allowed_oauth_flows_user_pool_client:
            params.update(AllowedOAuthFlowsUserPoolClient=
                          allowed_oauth_flows_user_pool_client)
        if analytics_configuration:
            params.update(AnalyticsConfiguration=analytics_configuration)
        if prevent_user_existence_errors:
            params.update(
                PreventUserExistenceErrors=prevent_user_existence_errors)
        if enable_token_revocation:
            params.update(EnableTokenRevocation=enable_token_revocation)

        response = self.client.create_user_pool_client(**params)
        return response['UserPoolClient'].get('ClientId')

    def list_user_pool_clients(self, cup_id):

        response = self.client.list_user_pool_clients(
            UserPoolId=cup_id,
            MaxResults=5
        )

        client_ids = [cup_client['ClientId'] for
                      cup_client in response['UserPoolClients']]
        next_token = response.get('NextToken')

        while next_token:
            response = self.client.list_user_pool_clients(
                UserPoolId=cup_id,
                MaxResults=5,
                NextToken=next_token
            )
            client_ids.extend([cup_client['ClientId'] for
                               cup_client in response['UserPoolClients']])
            next_token = response.get('NextToken')

        return client_ids

    def if_cup_client_exist(self, cup_name):

        cup_id = self.if_pool_exists_by_name(cup_name)
        if not cup_id:
            return

        client_ids = self.list_user_pool_clients(cup_id)

        if len(client_ids) == 1:
            return client_ids[0]
        if len(client_ids) > 1:
            _LOG.warn(f'Client ID of Cognito User Pool "{cup_name}" can\'t be '
                      f'identified unambiguously because there is more than '
                      f'one client in the Cognito User Pool. Determined IDs: '
                      f'"{client_ids}"')
        else:
            _LOG.warn(f'Clients not found in the Cognito User Pool with the '
                      f'name "{cup_name}"')

    def if_pool_exists_by_name(self, user_pool_name):
        ids = []
        paginator = self.client.get_paginator('list_user_pools')
        response = paginator.paginate(
            PaginationConfig={
                'MaxItems': 60,
                'PageSize': 10
            }
        )
        for page in response:
            ids.extend(
                [user_pool['Id'] for user_pool in page['UserPools'] if
                 user_pool['Name'] == user_pool_name]
            )
        next_token = response.resume_token
        while next_token:
            response = paginator.paginate(
                PaginationConfig={
                    'MaxItems': 60,
                    'PageSize': 10,
                    'StartingToken': next_token
                }
            )
            for page in response:
                ids.extend(
                    [user_pool['Id'] for user_pool in page['UserPools'] if
                     user_pool['Name'] == user_pool_name]
                )
            next_token = response.resume_token

        if len(ids) == 1:
            return ids[0]
        if len(ids) > 1:
            _LOG.warn(f'Cognito User Pool can\'t be identified unambiguously '
                      f'because there is more than one resource with the name '
                      f'"{user_pool_name}" in the region {self.region}. '
                      f'Determined IDs: "{ids}"')
        else:
            _LOG.warn(f'Cognito User Pool with the name "{user_pool_name}" '
                      f'not found in the region {self.region}')

    def describe_user_pool(self, user_pool_id):
        return self.client.describe_user_pool(UserPoolId=user_pool_id)

    def remove_user_pool(self, user_pool_id):
        """
        Removes user pool by id.

        :type user_pool_id: str
        """
        self.client.delete_user_pool(UserPoolId=user_pool_id)

    def add_custom_attributes(self, user_pool_id, attributes):
        self.client.add_custom_attributes(UserPoolId=user_pool_id,
                                          CustomAttributes=attributes)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import time

from syndicate.commons.log_helper import get_logger
from syndicate.core import ClientError
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.kinesis_resource')


class KinesisResource(BaseResource):

    def __init__(self, kin_conn) -> None:
        self.kin_conn = kin_conn

    def create_kinesis_stream(self, args):
        return self.create_pool(self._create_kinesis_stream_from_meta, args)

    def remove_kinesis_streams(self, args):
        self.create_pool(self._remove_kinesis_stream, args)

    @unpack_kwargs
    def _remove_kinesis_stream(self, arn, config):
        stream_name = config['resource_name']
        try:
            self.kin_conn.remove_stream(stream_name=stream_name)
            _LOG.info('Kinesis stream %s was removed.', stream_name)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn('Kinesis stream %s is not found', stream_name)
            else:
                raise e

    @unpack_kwargs
    def _create_kinesis_stream_from_meta(self, name, meta):
        response = self.kin_conn.get_stream(name)
        if response:
            stream_status = response['StreamDescription']['StreamStatus']
            if stream_status == 'DELETING':
                _LOG.debug('Waiting for deletion kinesis stream %s...', name)
                time.sleep(120)
            else:
                _LOG.warn('%s kinesis stream exists', name)
                return {
                    response['StreamARN']: build_description_obj(response,
                                                                 name, meta)
                }
        self.kin_conn.create_stream(stream_name=name,
                                    shard_count=meta['shard_count'])
        _LOG.info('Created kinesis stream %s.', name)
        return self.describe_kinesis_stream(name=name, meta=meta)

    def describe_kinesis_stream(self, name, meta):
        response = self.kin_conn.get_stream(name)
        return {
            response['StreamARN']: build_description_obj(response, name, meta)
        }



from abc import ABC, abstractmethod
from pprint import pformat

from syndicate.core.resources.helper import filter_dict_by_shape


class AbstractExternalResource(ABC):
    @abstractmethod
    def define_resource_shape(self):
        pass

    @abstractmethod
    def describe_meta(self, name):
        pass

    def compare_meta(self, name, syndicate_meta):
        aws_meta = self.describe_meta(name)

        resource_shape = self.define_resource_shape()
        syndicate_meta = self.filter_meta(
            meta=syndicate_meta,
            shape=resource_shape
        )

        aws_resource_meta = aws_meta.get(name)
        if not aws_resource_meta:
            return f"External resource '{name}' does not exist"

        aws_meta = self.filter_meta(
            meta=aws_resource_meta,
            shape=resource_shape
        )

        for key in syndicate_meta.keys():
            syndicate_value = syndicate_meta.get(key)
            aws_value = aws_meta.get(key)

            if isinstance(aws_value, list):
                try:  # sorting flat list
                    syndicate_value.sort()
                    aws_value.sort()
                except TypeError:  # sorting list of dicts
                    sort_key = list(aws_value[0].keys())[0]

                    syndicate_value.sort(key=lambda k: k[sort_key])
                    aws_value.sort(key=lambda k: k[sort_key])

        if syndicate_meta != aws_meta:
            return self.get_errors(
                resource_name=name,
                syndicate_meta=syndicate_meta,
                aws_meta=aws_meta
            )

    @staticmethod
    def filter_meta(meta, shape):
        return filter_dict_by_shape(meta, shape)

    @staticmethod
    def get_errors(resource_name, syndicate_meta, aws_meta):
        errors = [f"'{resource_name}' resource meta mismatch:"]
        for key in syndicate_meta.keys():
            syndicate_value = syndicate_meta.get(key)
            aws_value = aws_meta.get(key)

            if isinstance(aws_value, (str, int)) and syndicate_value != aws_value:
                errors.append(f"Expected '{key}' value: '{pformat(syndicate_value)}',\nGot '{pformat(aws_value)}'")
            if isinstance(aws_value, list):
                for aws_item, syndicate_item in zip(aws_value, syndicate_value):
                    if aws_item != syndicate_item:
                        errors.append(
                            f"Expected '{key}' value: '{pformat(syndicate_item)}',\nGot '{pformat(aws_item)}'")
        return '\n'.join(errors)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import glob
import json
import os
import platform
import shutil
import subprocess
import sys
from concurrent.futures import FIRST_EXCEPTION
from concurrent.futures import FIRST_EXCEPTION
from concurrent.futures.thread import ThreadPoolExecutor
from itertools import chain
from pathlib import Path
from typing import Union, Optional, List, Set

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.helper import build_py_package_name, zip_dir
from syndicate.core.conf.processor import path_resolver
from syndicate.core.constants import (LAMBDA_CONFIG_FILE_NAME, DEFAULT_SEP,
                                      REQ_FILE_NAME, LOCAL_REQ_FILE_NAME,
                                      LAMBDA_LAYER_CONFIG_FILE_NAME,
                                      PYTHON_LAMBDA_LAYER_PATH,
                                      MANY_LINUX_2014_PLATFORM)
from syndicate.core.helper import (build_path, unpack_kwargs, zip_ext,
                                   without_zip_ext)
from syndicate.core.resources.helper import validate_params

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

_PY_EXT = "*.py"
EMPTY_LINE_CHARS = ('\n', '\r\n')


def assemble_python_lambdas(project_path, bundles_dir):
    from syndicate.core import CONFIG
    project_base_folder = os.path.basename(os.path.normpath(project_path))
    if project_path != '.':
        project_abs_path = build_path(CONFIG.project_path, project_base_folder)
    else:
        project_abs_path = CONFIG.project_path
    _LOG.info('Going to process python project by path: {0}'.format(
        project_abs_path))
    with ThreadPoolExecutor(max_workers=5) as executor:
        futures = []
        for root, sub_dirs, files in os.walk(project_abs_path):
            for item in files:
                if item.endswith(LAMBDA_CONFIG_FILE_NAME):
                    _LOG.info('Going to build artifact in: {0}'.format(root))
                    arg = {
                        'root': str(Path(root)),
                        'config_file': str(Path(root, item)),
                        'target_folder': bundles_dir,
                        'project_path': project_path,
                    }
                    futures.append(
                        executor.submit(_build_python_artifact, arg))
                elif item.endswith(LAMBDA_LAYER_CONFIG_FILE_NAME):
                    _LOG.info(f'Going to build lambda layer in `{root}`')
                    arg = {
                        'layer_root': root,
                        'bundle_dir': bundles_dir,
                        'project_path': project_path
                    }
                    futures.append(
                        executor.submit(build_python_lambda_layer, arg))
        result = concurrent.futures.wait(futures, return_when=FIRST_EXCEPTION)
    for future in result.done:
        exception = future.exception()
        if exception:
            print(f'\033[91m' + str(exception), file=sys.stderr)
            sys.exit(1)
    _LOG.info('Python project was processed successfully')


def remove_dir(path: Union[str, Path]):
    removed = False
    while not removed:
        _LOG.info(f'Trying to remove "{path}"')
        try:
            shutil.rmtree(path)
            removed = True
        except Exception as e:
            _LOG.warn(f'An error "{e}" occurred while '
                      f'removing artifacts "{path}"')


@unpack_kwargs
def build_python_lambda_layer(layer_root: str, bundle_dir: str,
                              project_path: str):
    """
    Layer root is a dir where these files exist:
    - lambda_layer_config.json
    - local_requirements.txt
    - requirements.txt
    """
    with open(Path(layer_root, LAMBDA_LAYER_CONFIG_FILE_NAME), 'r') as file:
        layer_config = json.load(file)
    validate_params(layer_root, layer_config, ['name', 'deployment_package'])
    artifact_name = without_zip_ext(layer_config['deployment_package'])
    artifact_path = Path(bundle_dir, artifact_name)
    path_for_requirements = artifact_path / PYTHON_LAMBDA_LAYER_PATH
    _LOG.info(f'Artifacts path: {artifact_path}')
    os.makedirs(artifact_path, exist_ok=True)

    # install requirements.txt content
    requirements_path = Path(layer_root, REQ_FILE_NAME)
    if os.path.exists(requirements_path):
        install_requirements_to(requirements_path, to=path_for_requirements,
                                config=layer_config)

    # install local requirements
    local_requirements_path = Path(layer_root, LOCAL_REQ_FILE_NAME)
    if os.path.exists(local_requirements_path):
        _LOG.info('Going to install local dependencies')
        _install_local_req(path_for_requirements, local_requirements_path,
                           project_path)
        _LOG.info('Local dependencies were installed successfully')

    # making zip archive
    package_name = zip_ext(layer_config['deployment_package'])
    _LOG.info(f'Packaging artifacts by {artifact_path} to {package_name}')
    zip_dir(str(artifact_path), str(Path(bundle_dir, package_name)))
    _LOG.info(f'Package \'{package_name}\' was successfully created')
    # remove unused folder
    remove_dir(artifact_path)
    _LOG.info(f'"{artifact_path}" was removed successfully')


@unpack_kwargs
def _build_python_artifact(root, config_file, target_folder, project_path):
    _LOG.info(f'Building artifact in {target_folder}')

    # create folder to store artifacts
    with open(config_file, 'r') as file:
        lambda_config = json.load(file)
    validate_params(root, lambda_config, ['lambda_path', 'name', 'version'])
    artifact_name = f'{lambda_config["name"]}-{lambda_config["version"]}'
    artifact_path = Path(target_folder, artifact_name)
    _LOG.info(f'Artifacts path: {artifact_path}')
    os.makedirs(artifact_path, exist_ok=True)

    # install requirements.txt content
    requirements_path = Path(root, REQ_FILE_NAME)
    if os.path.exists(requirements_path):
        install_requirements_to(requirements_path, to=artifact_path,
                                config=lambda_config)

    # install local requirements
    local_requirements_path = Path(root, LOCAL_REQ_FILE_NAME)
    if os.path.exists(local_requirements_path):
        _LOG.info('Going to install local dependencies')
        _install_local_req(artifact_path, local_requirements_path,
                           project_path)
        _LOG.info('Local dependencies were installed successfully')

    # copy lambda's specific packages
    packages_dir = artifact_path / 'lambdas' / Path(root).name
    os.makedirs(packages_dir, exist_ok=True)
    for package in filter(
            is_python_package,
            [Path(root, item) for item in os.listdir(root)]):
        _LOG.info(f'Copying package {package} to lambda\'s artifacts packages '
                  f'dir: {packages_dir}')
        shutil.copytree(str(package), str(packages_dir / package.name))
        _LOG.info('Copied successfully')

    # copy lambda's handler to artifacts folder
    _LOG.info(f'Copying lambda\'s handler from {root} to {artifact_path}')
    _copy_py_files(root, artifact_path)

    # making zip archive
    package_name = build_py_package_name(lambda_config["name"],
                                         lambda_config["version"])
    _LOG.info(f'Packaging artifacts by {artifact_path} to {package_name}')
    zip_dir(str(artifact_path), str(Path(target_folder, package_name)))
    _LOG.info(f'Package \'{package_name}\' was successfully created')

    # remove unused folder
    remove_dir(artifact_path)
    _LOG.info(f'"{artifact_path}" was removed successfully')


def install_requirements_to(requirements_txt: Union[str, Path],
                            to: Union[str, Path],
                            config: Optional[dict] = None):
    config = config or {}
    _LOG.info('Going to install 3-rd party dependencies')
    supported_platforms = update_platforms(set(config.get('platforms') or []))
    python_version = _get_python_version(lambda_config=config)
    try:
        if supported_platforms:
            # tries to install packages compatible with specific platforms
            # returns the list of requirement that failed the installation
            failed_requirements = install_requirements_for_platform(
                requirements_txt=requirements_txt,
                to=to,
                supported_platforms=supported_platforms,
                python_version=python_version
            )
            for failed in failed_requirements:
                command = build_pip_install_command(  # default installation
                    requirement=failed,
                    to=to,
                )
                subprocess.run(command, stderr=subprocess.PIPE, check=True)
        else:
            _LOG.info('Installing all the requirements with defaults')
            command = build_pip_install_command(
                requirement=requirements_txt,
                to=to
            )
            subprocess.run(command, stderr=subprocess.PIPE, check=True)

    except subprocess.CalledProcessError as e:
        message = f'An error: \n"{e.stderr.decode()}"\noccured while ' \
                  f'installing requirements: "{str(requirements_txt)}" ' \
                  f'for package "{to}"'
        _LOG.error(message)
        raise RuntimeError(message)
    _LOG.info('3-rd party dependencies were installed successfully')


def build_pip_install_command(
        requirement: Optional[Union[str, Path]],
        to: Optional[Union[str, Path]] = None,
        implementation: Optional[str] = None,
        python: Optional[str] = None,
        only_binary: Optional[str] = None,
        platforms: Optional[Set[str]] = None,
        additional_args: Optional[List[str]] = None) -> List[str]:
    """
    :param requirement: path to requirements.txt or just one requirement.
    If the path is not file or does not exist it will be treated as one
    requirement.
    :param to: Optional[str] the path where to install requirements
    :param implementation: Optional[str], can be `cp`
    :param python: Optional[str], can be `3.8`, `3.9`
    :param only_binary: Optional[str], can be `:all:` or `:none:`
    :param platforms: Optional[Set], can be {'manylinux2014_x86_64'}
    :param additional_args: Optional[List[str]] list or some additional args
    :return: List[str]
    """
    command = [
        sys.executable, '-m', 'pip', 'install'
    ]
    r_path = Path(requirement)
    if r_path.exists() and r_path.is_file():
        command.extend(['-r', str(r_path)])
    else:  # not a path to requirements.txt but one requirement
        command.append(requirement)

    if to:
        command.extend(['-t', str(to)])
    if implementation:
        command.extend(['--implementation', 'cp'])
    if python:
        command.extend(['--python-version', python])
    if only_binary:
        command.append(f'--only-binary={only_binary}')
    if platforms:
        command.extend(chain.from_iterable(
            ('--platform', p) for p in platforms
        ))
    command.extend(additional_args or [])
    return command


def update_platforms(platforms: Set[str]) -> Set[str]:
    """
    If platforms are not empty, just return them without changing
    (user's choice). But in case the set is empty and the current
    processor is ARM (mac m1) or OS is Windows, we add
    manylinux2014_x86_64 to the list of platforms because by default
    lambdas with x86_64.
    This code is experimental and can be adjusted
    """
    if platforms:
        return platforms
    _arm = platform.processor() == 'arm'
    _win = platform.system() == 'Windows'
    if _arm or _win:
        platforms.add(MANY_LINUX_2014_PLATFORM)
    return platforms


def install_requirements_for_platform(requirements_txt: Union[str, Path],
                                      to: Union[str, Path],
                                      python_version: str,
                                      supported_platforms: Set[str]
                                      ) -> List[str]:
    _LOG.info(f'Going to install 3-rd party dependencies for platforms: '
              f'{",".join(supported_platforms)}')
    fp = open(requirements_txt, 'r')
    it = (
        line.split(' #')[0].strip() for line in
        filter(lambda line: not line.strip().startswith('#')
                            and line not in EMPTY_LINE_CHARS, fp)
    )
    failed_requirements = []
    for requirement in it:
        try:
            command = build_pip_install_command(
                requirement=requirement,
                to=to,
                implementation='cp',
                python=python_version,
                only_binary=':all:',
                platforms=supported_platforms
            )
            subprocess.run(command, stderr=subprocess.PIPE, check=True)
        except subprocess.CalledProcessError as e:
            message = f'An error: \n"{e.stderr.decode()}"\noccured while ' \
                      f'installing requirements for platforms:' \
                      f'{",".join(supported_platforms)}: ' \
                      f'"{str(requirements_txt)}" for package "{requirement}"'
            USER_LOG.warning(f"\033[93m{message}\033[0m")
            failed_requirements.append(requirement)
    fp.close()
    return failed_requirements


def _install_local_req(artifact_path, local_req_path, project_path):
    from syndicate.core import CONFIG
    with open(local_req_path) as f:
        local_req_list = f.readlines()
    local_req_list = [path_resolver(r.strip()) for r in local_req_list]
    _LOG.info(f'Installing local dependencies: {local_req_list}')
    # copy folders
    for lrp in local_req_list:
        _LOG.info(f'Processing local dependency: {lrp}')
        shutil.copytree(str(Path(CONFIG.project_path, project_path, lrp)),
                        str(Path(artifact_path, lrp)))
        _LOG.debug('Dependency was copied successfully')

        folders = [r for r in lrp.split(DEFAULT_SEP) if r]
        # process folder from root python project
        folders.insert(0, '')
        i = 0
        temp_path = ''
        while i < len(folders):
            temp_path += DEFAULT_SEP + folders[i]
            src_path = Path(CONFIG.project_path, project_path, temp_path)
            dst_path = Path(artifact_path, temp_path)
            _copy_py_files(str(src_path), str(dst_path))
            i += 1
        _LOG.debug('Python files from packages were copied successfully')


def _get_python_version(lambda_config: dict) -> Optional[str]:
    """
    Lambda config or layer config.
     "runtime": "python3.7" => "3.7".
    If "runtime" contains a list with runtimes. The lowest version is returned
    """
    runtimes: Union[None, List, str] = lambda_config.get('runtime')
    if not runtimes:
        return
    if isinstance(runtimes, str):
        runtimes = [runtimes]
    return sorted(
        ''.join(ch for ch in runtime if ch.isdigit() or ch == '.')
        for runtime in runtimes
    )[0]


def _copy_py_files(search_path, destination_path):
    files = glob.iglob(build_path(search_path, _PY_EXT))
    for py_file in files:
        if os.path.isfile(py_file):
            shutil.copy2(py_file, destination_path)


def is_python_package(path: Union[Path, str]) -> bool:
    """A file is considered to be a package if it's a directory containing
    __init__.py"""
    return os.path.isdir(path) and os.path.exists(Path(path, '__init__.py'))



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import (
    unpack_kwargs)
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.documentdb_resource')


class DocumentDBClusterResource(BaseResource):

    def __init__(self, docdb_conn, account_id,
                 region) -> None:
        self.connection = docdb_conn
        self.account_id = account_id
        self.region = region

    def create_db_cluster(self, args):
        """ Create DocumentDB cluster in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_db_cluster_from_meta,
                                args)

    def describe_documentdb_cluster(self, identifier, meta):
        if not identifier:
            return
        response = self.connection.describe_db_clusters(identifier)
        if not response:
            return
        arn = f'arn:aws:rds:{self.region}:{self.account_id}:' \
              f'cluster:{identifier}'
        return {
            arn: build_description_obj(response, identifier, meta)
        }

    @unpack_kwargs
    def _create_db_cluster_from_meta(self, name, meta):
        pool_id = self.connection.describe_db_clusters(name)
        if pool_id:
            _LOG.warn(f'\'{name}\' cluster exists.')
            return self.describe_documentdb_cluster(identifier=name, meta=meta)

        _LOG.info(f'Creating documentDB cluster {name}')
        vpc_security_group_ids = meta.get('vpc_security_group_ids', None)
        if vpc_security_group_ids and not isinstance(
                vpc_security_group_ids, list):
            _LOG.warn(f'Incorrect type for vpc_security_group_ids: '
                      f'{vpc_security_group_ids}')
            vpc_security_group_ids = None

        availability_zones = meta.get('availability_zones', None)
        if availability_zones and not isinstance(
                availability_zones, list):
            _LOG.warn(f'Incorrect type for availability_zones: '
                      f'{availability_zones}')
            availability_zones = None
        port = meta.get('port', None)
        master_username = meta.get('master_username', None)
        if master_username and master_username[0].isdigit():
            raise AssertionError(f'The first character of master username '
                                 f'must be a letter: {master_username}')
        master_password = meta.get('master_password', None)
        if master_password and any(
                char in master_password for char in ('"', '@', '/')):
            raise AssertionError(
                f'The password cannot contain forward slash (/), double quote '
                f'(") or the "at" symbol (@): {master_password}')
        cluster = self.connection.create_db_cluster(
            identifier=name, availability_zones=availability_zones,
            vpc_security_group_ids=vpc_security_group_ids, port=port,
            master_password=master_password, master_username=master_username)
        _LOG.info(f'Created documentDB cluster {cluster}')
        return self.describe_documentdb_cluster(identifier=name, meta=meta)

    def remove_db_cluster(self, args):
        self.create_pool(self._remove_db_cluster, args)

    @unpack_kwargs
    def _remove_db_cluster(self, arn, config):
        cluster = config['description']
        try:
            self.connection.delete_db_cluster(cluster)
            _LOG.info(f'DocumentDB cluster \'{cluster}\' was removed')
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn(f'DocumentDB cluster \'{cluster}\' is not found')
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os
import sys
from functools import partial

import click
from tabulate import tabulate

from syndicate.core.export.export_processor import export_specification
from syndicate.core.transform.transform_processor import generate_build_meta
from syndicate.core import initialize_connection, \
    initialize_project_state, initialize_signal_handling
from syndicate.core.build.artifact_processor import (RUNTIME_NODEJS,
                                                     assemble_artifacts,
                                                     RUNTIME_JAVA,
                                                     RUNTIME_PYTHON,
                                                     RUNTIME_SWAGGER_UI)
from syndicate.core.build.bundle_processor import (create_bundles_bucket,
                                                   load_bundle,
                                                   upload_bundle_to_s3,
                                                   if_bundle_exist)
from syndicate.core.build.deployment_processor import (
    continue_deployment_resources, create_deployment_resources,
    remove_deployment_resources, update_deployment_resources)
from syndicate.core.build.meta_processor import create_meta
from syndicate.core.build.profiler_processor import (get_metric_statistics,
                                                     process_metrics)
from syndicate.core.build.warmup_processor import (process_deploy_resources,
                                                   process_api_gw_resources,
                                                   warm_upper,
                                                   process_existing_api_gw_id,
                                                   process_inputted_api_gw_id)
from syndicate.core.conf.validator import (JAVA_LANGUAGE_NAME,
                                           PYTHON_LANGUAGE_NAME,
                                           NODEJS_LANGUAGE_NAME,
                                           SWAGGER_UI_NAME)
from syndicate.core.decorators import (check_deploy_name_for_duplicates,
                                       check_deploy_bucket_exists)
from syndicate.core.groups.generate import (generate,
                                            GENERATE_PROJECT_COMMAND_NAME,
                                            GENERATE_CONFIG_COMMAND_NAME)
from syndicate.core.groups.tags import tags
from syndicate.core.helper import (create_bundle_callback,
                                   handle_futures_progress_bar,
                                   resolve_path_callback, timeit,
                                   verify_bundle_callback, sync_lock,
                                   resolve_default_value, ValidRegionParamType,
                                   generate_default_bundle_name,
                                   resolve_and_verify_bundle_callback,
                                   param_to_lower, verbose_option,
                                   validate_incompatible_options)
from syndicate.core.project_state.project_state import (MODIFICATION_LOCK,
                                                        WARMUP_LOCK)
from syndicate.core.project_state.status_processor import project_state_status
from syndicate.core.project_state.sync_processor import sync_project_state
from syndicate.core.constants import TEST_ACTION, BUILD_ACTION, \
    DEPLOY_ACTION, UPDATE_ACTION, CLEAN_ACTION, SYNC_ACTION, \
    STATUS_ACTION, WARMUP_ACTION, PROFILER_ACTION, ASSEMBLE_JAVA_MVN_ACTION, \
    ASSEMBLE_PYTHON_ACTION, ASSEMBLE_NODE_ACTION, ASSEMBLE_ACTION, \
    PACKAGE_META_ACTION, CREATE_DEPLOY_TARGET_BUCKET_ACTION, UPLOAD_ACTION, \
    COPY_BUNDLE_ACTION, EXPORT_ACTION, ASSEMBLE_SWAGGER_UI_ACTION

INIT_COMMAND_NAME = 'init'
SYNDICATE_PACKAGE_NAME = 'aws-syndicate'
HELP_PARAMETER_KEY = '--help'
commands_without_config = (
    INIT_COMMAND_NAME,
    GENERATE_PROJECT_COMMAND_NAME,
    GENERATE_CONFIG_COMMAND_NAME,
    HELP_PARAMETER_KEY
)


def _not_require_config(all_params):
    return any(item in commands_without_config for item in all_params)


@click.group(name='syndicate')
@click.version_option()
def syndicate():
    from syndicate.core import CONF_PATH
    if _not_require_config(sys.argv):
        pass
    elif CONF_PATH:
        click.echo('Configuration used: ' + CONF_PATH)
        initialize_connection()
        initialize_project_state()
        initialize_signal_handling()
    else:
        click.echo('Environment variable SDCT_CONF is not set! '
                   'Please verify that you have provided path to '
                   'correct config files '
                   'or execute `syndicate generate config` command.')
        sys.exit(1)


@syndicate.command(name=TEST_ACTION)
@click.option('--suite', default='unittest',
              type=click.Choice(['unittest', 'pytest', 'nose'],
                                case_sensitive=False),
              help='Supported testing frameworks. Possible options: unittest, '
                   'pytest, nose. Default value: unittest')
@click.option('--test_folder_name', nargs=1, default='tests',
              help='Directory in the project that contains tests to run. '
                   'Default folder: tests')
@click.option('--errors_allowed', is_flag=True,
              help='Flag to return successful response even if some tests '
                   'fail')
@verbose_option
@timeit(action_name=TEST_ACTION)
def test(suite, test_folder_name, errors_allowed):
    """Discovers and runs tests inside python project configuration path."""
    click.echo('Running tests...')
    import subprocess
    from syndicate.core import CONFIG
    project_path = CONFIG.project_path
    test_folder = os.path.join(project_path, test_folder_name)
    if not os.path.exists(test_folder):
        click.echo(f'Tests not found, \'{test_folder_name}\' folder is missing'
                   f' in \'{project_path}\'.')
        return
    test_lib_command_mapping = {
        'unittest': f'{sys.executable} -m unittest discover {test_folder} -v',
        'pytest': 'pytest --no-header -v',
        'nose': 'nosetests --verbose'
    }

    command = test_lib_command_mapping.get(suite)
    result = subprocess.run(command.split(), cwd=project_path)

    if not errors_allowed:
        if result.returncode != 0:
            click.echo('Some tests failed. Exiting.')
            sys.exit(1)
    click.echo('Tests passed.')


@syndicate.command(name=BUILD_ACTION)
@click.option('--bundle_name', '-b', nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle to build. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%SZ')
@click.option('--force_upload', '-F', is_flag=True, default=False,
              help='Flag to override existing bundle with the same name')
@click.option('--errors_allowed', is_flag=True, default=False,
              help='Flag to continue bundle building if some tests fail')
@verbose_option
@click.pass_context
@check_deploy_bucket_exists
@timeit(action_name=BUILD_ACTION)
def build(ctx, bundle_name, force_upload, errors_allowed):
    """
    Builds bundle of an application
    """
    if if_bundle_exist(bundle_name=bundle_name) and not force_upload:
        click.echo(f'Bundle name \'{bundle_name}\' already exists '
                   f'in deploy bucket. Please use another bundle '
                   f'name or delete the bundle')
        return
    ctx.invoke(test, errors_allowed=errors_allowed)
    ctx.invoke(assemble, bundle_name=bundle_name, force_upload=force_upload)
    ctx.invoke(package_meta, bundle_name=bundle_name)
    ctx.invoke(upload, bundle_name=bundle_name, force_upload=force_upload)


@syndicate.command(name='transform')
@click.option('--bundle_name',
              callback=resolve_default_value,
              help='Name of the bundle to transform. '
                   'Default value: name of the latest built bundle')
@click.option('--dsl', type=click.Choice(['CloudFormation', 'Terraform'],
                                         case_sensitive=False),
              callback=param_to_lower,
              multiple=True, required=True,
              help='Type of the IaC provider')
@click.option('--output_dir',
              help='The directory where a transformed template will be saved')
@verbose_option
@timeit()
def transform(bundle_name, dsl, output_dir):
    """
    Transforms the meta-description of a bundle to a template
    compatible with the specified IaC provider
    """
    generate_build_meta(bundle_name=bundle_name,
                        dsl_list=dsl,
                        output_directory=output_dir)


@syndicate.command(name=DEPLOY_ACTION)
@sync_lock(lock_type=MODIFICATION_LOCK)
@click.option('--deploy_name', '-d', callback=resolve_default_value,
              help='Name of the deploy. Default value: name of the project')
@click.option('--bundle_name', '-b', callback=resolve_default_value,
              help='Name of the bundle to deploy. '
                   'Default value: name of the latest built bundle')
@click.option('--deploy_only_types', '-types', multiple=True,
              help='Types of the resources to deploy')
@click.option('--deploy_only_resources', '-resources', multiple=True,
              help='Names of the resources to deploy')
@click.option('--deploy_only_resources_path', '-path', nargs=1,
              help='Path to file containing names of the resources to deploy')
@click.option('--excluded_resources', '-exresources', multiple=True,
              help='Names of the resources to skip while deploy.')
@click.option('--excluded_resources_path', '-expath', nargs=1,
              help='Path to file containing names of the resources to skip '
                   'while deploy')
@click.option('--excluded_types', '-extypes', multiple=True,
              help='Types of the resources to skip while deploy')
@click.option('--continue_deploy', is_flag=True, is_eager=True,
              help='Flag to continue failed deploy')
@click.option('--replace_output', is_flag=True, default=False,
              help='Flag to replace the existing deploy output')
@click.option('--rollback_on_error', is_flag=True, default=False,
              callback=partial(validate_incompatible_options,
                               incompatible_options=['continue_deploy']),
              help='Flag to automatically clean deployed resources if the'
                   ' deployment is unsuccessful. Cannot be used with'
                   ' --continue_deploy flag.')
@verbose_option
@check_deploy_name_for_duplicates
@check_deploy_bucket_exists
@timeit(action_name=DEPLOY_ACTION)
def deploy(deploy_name, bundle_name, deploy_only_types, deploy_only_resources,
           deploy_only_resources_path, excluded_resources,
           excluded_resources_path, excluded_types, continue_deploy,
           replace_output, rollback_on_error):
    """
    Deploys the application infrastructure
    """
    if deploy_only_resources_path and os.path.exists(
            deploy_only_resources_path):
        deploy_resources_list = json.load(open(deploy_only_resources_path))
        deploy_only_resources = tuple(
            set(deploy_only_resources + tuple(deploy_resources_list)))
        if deploy_only_resources:
            click.echo(
                f'Resources to update: {list(deploy_only_resources)}')

    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
        if excluded_resources:
            click.echo(
                f'Resources to update: {list(excluded_resources)}')

    if continue_deploy:
        deploy_success = continue_deployment_resources(deploy_name,
                                                       bundle_name,
                                                       deploy_only_resources,
                                                       deploy_only_types,
                                                       excluded_resources,
                                                       excluded_types,
                                                       replace_output
                                                       )

    else:
        deploy_success = create_deployment_resources(deploy_name, bundle_name,
                                                     deploy_only_resources,
                                                     deploy_only_types,
                                                     excluded_resources,
                                                     excluded_types,
                                                     replace_output,
                                                     rollback_on_error
                                                     )

    message = 'Backend resources were deployed{suffix}.'.format(
        suffix='' if deploy_success else (
            ' with errors. Rollback is enabled, output file was not created'
            if rollback_on_error else ' with errors. See deploy output file'
        )
    )
    click.echo(message)
    return deploy_success


@syndicate.command(name=UPDATE_ACTION)
@sync_lock(lock_type=MODIFICATION_LOCK)
@click.option('--bundle_name', '-b', callback=resolve_default_value,
              help='Name of the bundle to deploy. '
                   'Default value: name of the latest built bundle')
@click.option('--deploy_name', '-d', callback=resolve_default_value,
              help='Name of the deploy. Default value: name of the project')
@click.option('--update_only_types', '-types', multiple=True,
              help='Types of the resources to update')
@click.option('--update_only_resources', '-resources', multiple=True,
              help='Names of the resources to update')
@click.option('--update_only_resources_path', '-path', nargs=1,
              help='Path to file containing names of the resources to skip '
                   'while deploy')
@click.option('--excluded_resources', '-exresources', multiple=True,
              help='Names of the resources to skip while update.')
@click.option('--excluded_resources_path', '-expath', nargs=1,
              help='Path to file containing names of the resources to skip '
                   'while update')
@click.option('--excluded_types', '-extypes', multiple=True,
              help='Types of the resources to skip while update')
@click.option('--replace_output', nargs=1, is_flag=True, default=False,
              help='The flag to replace the existing deploy output file')
@verbose_option
@check_deploy_name_for_duplicates
@check_deploy_bucket_exists
@timeit(action_name=UPDATE_ACTION)
def update(bundle_name, deploy_name, replace_output, update_only_resources,
           update_only_resources_path, update_only_types, excluded_resources,
           excluded_resources_path, excluded_types):
    """
    Updates infrastructure from the provided bundle
    """
    click.echo(f'Bundle name: {bundle_name}')

    if update_only_resources_path and os.path.exists(
            update_only_resources_path):
        update_resources_list = json.load(open(update_only_resources_path))
        update_only_resources = tuple(
            set(update_only_resources + tuple(update_resources_list)))
        if update_only_resources:
            click.echo(
                f'Resources to update: {list(update_only_resources)}')

    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
        if excluded_resources:
            click.echo(
                f'Resources excluded from update: {list(excluded_resources)}')

    success = update_deployment_resources(
        bundle_name=bundle_name,
        deploy_name=deploy_name,
        update_only_types=update_only_types,
        update_only_resources=update_only_resources,
        excluded_resources=excluded_resources,
        excluded_types=excluded_types,
        replace_output=replace_output)
    if success:
        click.echo('Update of resources has been successfully completed')
    else:
        click.echo('Something went wrong during resources update')


@syndicate.command(name=CLEAN_ACTION)
@sync_lock(lock_type=MODIFICATION_LOCK)
@click.option('--deploy_name', '-d', nargs=1, callback=resolve_default_value,
              help='Name of the deploy. This parameter allows the framework '
                   'to decide,which exactly output file should be used. The '
                   'resources are cleaned based on the output file which is '
                   'created during the deployment process. If not specified, '
                   'resolves the latest deploy name')
@click.option('--bundle_name', '-b', nargs=1, callback=resolve_default_value,
              help='Name of the bundle. If not specified, resolves the latest '
                   'bundle name')
@click.option('--clean_only_types', '-types', multiple=True,
              help='If specified only provided types will be cleaned')
@click.option('--clean_only_resources', '-resources', multiple=True,
              help='If specified only provided resources will be cleaned')
@click.option('--clean_only_resources_path', '-path', nargs=1, type=str,
              help='If specified only resources path will be cleaned')
@click.option('--clean_externals', nargs=1, is_flag=True, default=False,
              help='Flag. If specified only external resources will be '
                   'cleaned')
@click.option('--excluded_resources', '-exresources', multiple=True,
              help='If specified provided resources will be excluded')
@click.option('--excluded_resources_path', '-expath', nargs=1, type=str,
              help='If specified provided resource path will be excluded')
@click.option('--excluded_types', '-extypes', multiple=True,
              help='If specified provided types will be excluded')
@click.option('--preserve_state', is_flag=True,
              help='Preserve deploy output json file after resources removal')
@verbose_option
@timeit(action_name=CLEAN_ACTION)
def clean(deploy_name, bundle_name, clean_only_types, clean_only_resources,
          clean_only_resources_path, clean_externals, excluded_resources,
          excluded_resources_path, excluded_types, preserve_state):
    """
    Cleans the application infrastructure
    """
    click.echo('Command clean')
    click.echo(f'Deploy name: {deploy_name}')
    separator = ', '
    if clean_only_types:
        click.echo(f'Clean only types: {separator.join(clean_only_types)}')
    if clean_only_resources:
        click.echo(f'Clean only resources: '
                   f'{separator.join(clean_only_resources)}')
    if clean_only_resources_path:
        click.echo(f'Clean only resources path: {clean_only_resources_path}')
    if excluded_resources:
        click.echo(f'Excluded resources: {separator.join(excluded_resources)}')
    if excluded_resources_path:
        click.echo(f'Excluded resources path: {excluded_resources_path}')
    if excluded_types:
        click.echo(f'Excluded types: {separator.join(excluded_resources)}')
    if clean_only_resources_path and os.path.exists(clean_only_resources_path):
        clean_resources_list = json.load(open(clean_only_resources_path))
        clean_only_resources = tuple(
            set(clean_only_resources + tuple(clean_resources_list)))
    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))

    result = remove_deployment_resources(
        deploy_name=deploy_name,
        bundle_name=bundle_name,
        clean_only_resources=clean_only_resources,
        clean_only_types=clean_only_types,
        excluded_resources=excluded_resources,
        excluded_types=excluded_types,
        clean_externals=clean_externals,
        preserve_state=preserve_state)
    click.echo('AWS resources were removed.')
    return result


@syndicate.command(name=SYNC_ACTION)
@verbose_option
@check_deploy_bucket_exists
@timeit()
def sync():
    """
    Syncs the state of local project state file (.syndicate) and
    the remote one.
    """
    return sync_project_state()


@syndicate.command(name=STATUS_ACTION)
@click.option('--events', 'category', flag_value='events',
              help='Show event logs of the project')
@click.option('--resources', 'category', flag_value='resources',
              help='Show a summary of the project resources')
@verbose_option
@check_deploy_bucket_exists
@timeit()
def status(category):
    """
    Shows the state of a local project state file (.syndicate).
    Command displays the following content: project name, state, latest
    modification, locks summary, latest event, project resources.

    NOTE: The flags are incompatible and the status is displayed according to
    the first entered flag.
    """
    click.echo(project_state_status(category))


@syndicate.command(name=WARMUP_ACTION)
@sync_lock(lock_type=WARMUP_LOCK)
@click.option('--bundle_name', '-b', nargs=1, callback=resolve_default_value,
              help='Name of the bundle. If not specified, resolves the latest '
                   'bundle name')
@click.option('--deploy_name', '-d', nargs=1, callback=resolve_default_value,
              help='Name of the deploy. If not specified, resolves the latest '
                   'deploy name')
@click.option('--api_gw_id', '-api', nargs=1, multiple=True, type=str,
              help='Provide API Gateway IDs to warmup')
@click.option('--stage_name', '-stage', nargs=1, multiple=True, type=str,
              help='Name of stages of provided API Gateway IDs')
@click.option('--lambda_auth', '-auth', default=False, is_flag=True,
              help='Flag. Should be specified if API Gateway Lambda Authorizer'
                   ' is enabled')
@click.option('--header_name', '-hname', nargs=1,
              help='Name of authentication header.')
@click.option('--header_value', '-hvalue',
              nargs=1, help='Authentication header value.')
@verbose_option
@check_deploy_bucket_exists
@timeit(action_name=WARMUP_ACTION)
def warmup(bundle_name, deploy_name, api_gw_id, stage_name, lambda_auth,
           header_name, header_value):
    """
    Warmups Lambda functions
    """

    if bundle_name and deploy_name:
        click.echo(f'Deploy name: {deploy_name}')
        if not if_bundle_exist(bundle_name=bundle_name):
            click.echo(f'Bundle name \'{bundle_name}\' does not exists '
                       'in deploy bucket. Please use another bundle '
                       'name or create the bundle')
            return

        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            process_deploy_resources(deploy_name=deploy_name,
                                     bundle_name=bundle_name)

    elif api_gw_id:
        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            process_inputted_api_gw_id(api_id=api_gw_id, stage_name=stage_name,
                                       echo=click.echo)

    else:
        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            process_existing_api_gw_id(stage_name=stage_name, echo=click.echo)

    if not paths_to_be_triggered or not resource_path_warmup_key_mapping:
        click.echo('No resources to warm up')
        return
    resource_method_mapping, resource_warmup_key_mapping = \
        process_api_gw_resources(paths_to_be_triggered=paths_to_be_triggered,
                                 resource_path_warmup_key_mapping=
                                 resource_path_warmup_key_mapping)
    warm_upper(resource_method_mapping=resource_method_mapping,
               resource_warmup_key_mapping=resource_warmup_key_mapping,
               lambda_auth=lambda_auth, header_name=header_name,
               header_value=header_value)
    click.echo('Application resources have been warmed up.')


@syndicate.command(name=PROFILER_ACTION)
@click.option('--bundle_name', '-b', nargs=1, callback=resolve_default_value,
              help='The name of the bundle from which to select lambdas for '
                   'collecting metrics. If not specified, resolves the latest '
                   'bundle name')
@click.option('--deploy_name', '-d', nargs=1, callback=resolve_default_value,
              help='Name of the deploy. If not specified, resolves the latest '
                   'deploy name')
@click.option('--from_date', '-from', nargs=1,
              type=click.DateTime(formats=['%Y-%m-%dT%H:%M:%SZ']),
              help='Date from which collect lambda metrics. The '
                   '\'--to_date\' parameter required. Example of the date '
                   'format: 2022-02-02T02:02:02Z')
@click.option('--to_date', '-to', nargs=1,
              type=click.DateTime(formats=['%Y-%m-%dT%H:%M:%SZ']),
              help='Date until which collect lambda metrics. The '
                   '\'--from_date\' parameter required. Example of the date '
                   'format: 2022-02-02T02:02:02Z')
@verbose_option
@check_deploy_bucket_exists
def profiler(bundle_name, deploy_name, from_date, to_date):
    """
    Displays application Lambda metrics
    """
    metric_value_dict = get_metric_statistics(bundle_name, deploy_name,
                                              from_date, to_date)
    for lambda_name, metrics in metric_value_dict.items():
        prettify_metrics_dict = {}

        click.echo(f'{os.linesep}Lambda function name: {lambda_name}')
        prettify_metrics_dict = process_metrics(prettify_metrics_dict, metrics)
        if not prettify_metrics_dict:
            click.echo('No executions found')
        click.echo(tabulate(prettify_metrics_dict, headers='keys',
                            stralign='right'))


# =============================================================================


@syndicate.command(name=ASSEMBLE_JAVA_MVN_ACTION)
@timeit()
@click.option('--bundle_name', '-b', nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the Java project. The provided path is the '
                   'path for an mvn clean install. The artifacts are copied '
                   'to a folder, which is be later used as the deployment '
                   'bundle (the bundle path: bundles/${bundle_name})')
@click.option('--force_upload', '-fu', nargs=1,
              callback=resolve_path_callback, required=True,
              help='Identifier that indicates whether a locally existing'
                   ' bundle should be deleted and a new one created using'
                   ' the same path.')
@verbose_option
@timeit(action_name=ASSEMBLE_JAVA_MVN_ACTION)
def assemble_java_mvn(bundle_name, project_path, force_upload):
    """
    Builds Java lambdas

    \f
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :force_upload: force upload identification
    :return:
    """
    click.echo(f'Command compile java project path: {project_path}')
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_JAVA,
                       force_upload=force_upload
                       )
    click.echo('Java artifacts were prepared successfully.')


@syndicate.command(name=ASSEMBLE_PYTHON_ACTION)
@timeit()
@click.option('--bundle_name', '-b', nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the Python project. The code is '
                   'packed to a zip archive, where the external libraries are '
                   'found, which are described in the requirements.txt file, '
                   'and internal project dependencies according to the '
                   'described in local_requirements.txt file')
@click.option('--force_upload', '-fu', nargs=1,
              callback=resolve_path_callback, required=True,
              help='Identifier that indicates whether a locally existing'
                   ' bundle should be deleted and a new one created using'
                   ' the same path.')
@verbose_option
@timeit(action_name=ASSEMBLE_PYTHON_ACTION)
def assemble_python(bundle_name, project_path, force_upload):
    """
    Builds Python lambdas

    \f
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :force_upload: force upload identification
    :return:
    """
    click.echo(f'Command assemble python: project_path: {project_path} ')
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_PYTHON,
                       force_upload=force_upload
                       )
    click.echo('Python artifacts were prepared successfully.')


@syndicate.command(name=ASSEMBLE_NODE_ACTION)
@timeit()
@click.option('--bundle_name', '-b', nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the NodeJS project. The code is '
                   'packed to a zip archive, where the external libraries are '
                   'found, which are described in the package.json file')
@click.option('--force_upload', '-fu', nargs=1,
              callback=resolve_path_callback, required=True,
              help='Identifier that indicates whether a locally existing'
                   ' bundle should be deleted and a new one created using'
                   ' the same path.')
@verbose_option
@timeit(action_name=ASSEMBLE_NODE_ACTION)
def assemble_node(bundle_name, project_path, force_upload):
    """
    Builds NodeJS lambdas

    \f
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :force_upload: force upload identification
    :return:
    """
    click.echo(f'Command assemble node: project_path: {project_path} ')
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_NODEJS,
                       force_upload=force_upload
                       )
    click.echo('NodeJS artifacts were prepared successfully.')


@syndicate.command(name=ASSEMBLE_SWAGGER_UI_ACTION)
@timeit()
@click.option('--bundle_name', '-b', nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the project. Related files will be packed '
                   'into a zip archive.')
@verbose_option
@timeit(action_name=ASSEMBLE_SWAGGER_UI_ACTION)
def assemble_swagger_ui(**kwargs):
    """
        Builds Swagger UI artifacts

        \f
        :param bundle_name: name of the bundle
        :param project_path: path to project folder
        :return:
        """
    bundle_name = kwargs.get('bundle_name')
    project_path = kwargs.get('project_path')
    click.echo(f'Command assemble Swagger UI: project_path: {project_path} ')
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_SWAGGER_UI)
    click.echo('Swagger UI artifacts were prepared successfully.')


RUNTIME_LANG_TO_BUILD_MAPPING = {
    JAVA_LANGUAGE_NAME: assemble_java_mvn,
    PYTHON_LANGUAGE_NAME: assemble_python,
    NODEJS_LANGUAGE_NAME: assemble_node,
    SWAGGER_UI_NAME: assemble_swagger_ui
}


@syndicate.command(name=ASSEMBLE_ACTION)
@click.option('--bundle_name', '-b', callback=generate_default_bundle_name,
              help='Bundle\'s name to build the lambdas in. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--force_upload', '-fu', nargs=1,
              callback=resolve_path_callback, default=False,
              help='Identifier that indicates whether a locally existing'
                   ' bundle should be deleted and a new one created using'
                   ' the same path.')
@verbose_option
@click.pass_context
@timeit(action_name=ASSEMBLE_ACTION)
def assemble(ctx, bundle_name, force_upload):
    """
    Builds the application artifacts

    \f
    :param ctx:
    :param bundle_name: name of the bundle to which the artifacts
        will be associated
    :return:
    """
    click.echo(f'Building artifacts, bundle: {bundle_name}')
    from syndicate.core import PROJECT_STATE
    # Updates stale lambda state aggregation
    PROJECT_STATE.refresh_lambda_state()

    build_mapping_dict: dict = PROJECT_STATE.load_project_build_mapping()
    if build_mapping_dict:
        for key, value in build_mapping_dict.items():
            func = RUNTIME_LANG_TO_BUILD_MAPPING.get(key)
            if func:
                ctx.invoke(func, bundle_name=bundle_name,
                           project_path=value, force_upload=force_upload)
            else:
                click.echo(f'Build tool is not supported: {key}')
    else:
        click.echo('Projects to be built are not found')


@syndicate.command(name=PACKAGE_META_ACTION)
@click.option('--bundle_name', '-b', required=True,
              callback=verify_bundle_callback,
              help='Bundle\'s name to package the current meta in')
@verbose_option
@timeit(action_name=PACKAGE_META_ACTION)
def package_meta(bundle_name):
    """
    Generates metadata about the application infrastructure

    \f
    :param bundle_name: name of the bundle to generate metadata
    :return:
    """
    from syndicate.core import CONFIG
    click.echo(f'Package meta, bundle: {bundle_name}')
    create_meta(project_path=CONFIG.project_path,
                bundle_name=bundle_name)
    click.echo('Meta was configured successfully.')


@syndicate.command(name=CREATE_DEPLOY_TARGET_BUCKET_ACTION)
@verbose_option
@timeit()
def create_deploy_target_bucket():
    """
    Creates a bucket in AWS account where all bundles will be uploaded
    """
    from syndicate.core import CONFIG
    click.echo(f'Create deploy target sdk: {CONFIG.deploy_target_bucket}')
    create_bundles_bucket()
    click.echo('Deploy target bucket was created successfully')


@syndicate.command(name=UPLOAD_ACTION)
@click.option('--bundle_name', '-b',
              callback=resolve_and_verify_bundle_callback,
              help='Bundle name to which the build artifacts are gathered '
                   'and later used for the deployment. NOTE: if not '
                   'specified, the latest build will be uploaded')
@click.option('--force_upload', '-F', is_flag=True,
              help='Flag to override existing bundle with the same name as '
                   'provided')
@verbose_option
@check_deploy_bucket_exists
@timeit(action_name=UPLOAD_ACTION)
def upload(bundle_name, force_upload=False):
    """
    Uploads bundle from local storage to AWS S3

    \f
    :param bundle_name: name of the bundle to upload
    :param force_upload: used if the bundle with the same name as provided
        already exists in an account
    :return:
    """
    click.echo(f'Upload bundle: {bundle_name}')
    if force_upload:
        click.echo('Force upload')

    futures = upload_bundle_to_s3(bundle_name=bundle_name, force=force_upload)
    handle_futures_progress_bar(futures)

    click.echo('Bundle was uploaded successfully')


@syndicate.command(name=COPY_BUNDLE_ACTION)
@click.option('--bundle_name', '-b', nargs=1, callback=create_bundle_callback,
              required=True,
              help='The bundle name, to which the build artifacts '
                   'are gathered and later used for the deployment')
@click.option('--src_account_id', '-acc_id', nargs=1, required=True,
              help='The account ID, to which the bundle is to be '
                   'uploaded')
@click.option('--src_bucket_region', '-r', nargs=1, required=True,
              type=ValidRegionParamType(),
              help='The name of the region of the bucket where target bundle '
                   'is stored')
@click.option('--src_bucket_name', '-bucket', nargs=1, required=True,
              help='The name of the bucket where target bundle is stored')
@click.option('--role_name', '-role', nargs=1, required=True,
              help='The role name from the specified account, which is '
                   'assumed. Here you have to check the trusted relationship '
                   'between the accounts. The active account must be a trusted'
                   ' one for the account which is specified in the command')
@click.option('--force_upload', '-F', is_flag=True, default=False,
              help='Flag. Used if the bundle with the same name as provided '
                   'already exists in a target account')
@verbose_option
@timeit()
@click.pass_context
def copy_bundle(ctx, bundle_name, src_account_id, src_bucket_region,
                src_bucket_name, role_name, force_upload):
    """
    Copies the bundle from the specified account-region-bucket to
    account-region-bucket specified in syndicate.yml

    \f
    :param ctx:
    :param bundle_name: name of the bundle to copy
    :param src_account_id: id of the account where target bundle is stored
    :param src_bucket_region: region of the bucket where target bundle
        is stored
    :param src_bucket_name: name of the bucket where target bundle is stored
    :param role_name: name of the role that is assumed while copying
    :param force_upload: used if the bundle with the same name as provided
        already exists in a target account
    :return:
    """
    click.echo(f'Copy bundle: {bundle_name}')
    click.echo(f'Bundle name: {bundle_name}')
    click.echo(f'Source account id: {src_account_id}')
    click.echo(f'Source bucket region: {src_bucket_region}')
    click.echo(f'Source bucket name: {src_bucket_name}')
    futures = load_bundle(bundle_name, src_account_id, src_bucket_region,
                          src_bucket_name, role_name)
    handle_futures_progress_bar(futures)
    click.echo('Bundle was downloaded successfully')
    ctx.invoke(upload, bundle_name=bundle_name, force=force_upload)
    click.echo('Bundle was copied successfully')


@syndicate.command(name=EXPORT_ACTION)
@click.option('--resource_type',
              type=click.Choice(['api_gateway']), required=True,
              help='The type of resource to export configuration')
@click.option('--dsl',
              type=click.Choice(['oas_v3']), default='oas_v3',
              help='DSL of output specification')
@click.option('--deploy_name', '-d', nargs=1, callback=resolve_default_value,
              help='Name of the deploy. This parameter allows the framework '
                   'to decide,which exactly output file should be used. If '
                   'not specified, resolves the latest deploy name')
@click.option('--bundle_name',
              callback=resolve_default_value,
              help='Name of the bundle to export from. '
                   'Default value: name of the latest built bundle')
@click.option('--output_dir',
              help='The directory where an exported configuration will be '
                   'saved. If not specified, the directory with the name '
                   '"export" will be created in the project root directory to '
                   'store export files')
@verbose_option
def export(resource_type, dsl, deploy_name, bundle_name, output_dir):
    """
    Exports a configuration of the specified resource type to the file in a
    specified DSL

    param: resource_type: the type of the resource
    param: dsl: the DSL of the output configuration
    param: deploy_name: the name of the deployment
    param: bundle_name: the name of the bundle to export from
    param: output_dir: the directory where an exported specification will be
    saved
    """
    export_specification(deploy_name=deploy_name,
                         bundle_name=bundle_name,
                         output_directory=output_dir,
                         resource_type=resource_type,
                         dsl=dsl)
    if resource_type == 'api_gateway' and dsl == 'oas_v3':
        click.secho('Please note the AWS API Gateway-specific extensions are '
                    'used to define the API in OAS v3 that starting with '
                    '"x-amazon"', fg='yellow')


syndicate.add_command(generate)
syndicate.add_command(tags)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry
from syndicate.connection.iam_connection import IAMConnection

_LOG = get_logger('syndicate.connection.cognito_identity_connection')


@apply_methods_decorator(retry())
class CognitoIdentityConnection(object):
    """ Cognito identity connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region,
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = client('cognito-identity', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Cognito identity connection.')

    def create_identity_pool(self, pool_name, provider_name=None,
                             allow_unauthenticated=False, login_providers=None,
                             open_id_connect_provider_arns=None,
                             cognito_identity_providers=None,
                             saml_provider_arns=None):
        """ Crete Cognito identity pool and get identity pool id.

        :type pool_name: str
        :type provider_name: str ('login.provider.com')
        :type allow_unauthenticated: bool
        :type login_providers: dict
        :type open_id_connect_provider_arns: list
        :type cognito_identity_providers: list
        :type saml_provider_arns: list
        """
        params = dict(IdentityPoolName=pool_name,
                      AllowUnauthenticatedIdentities=allow_unauthenticated)
        if provider_name:
            params['DeveloperProviderName'] = provider_name
        if login_providers:
            params['SupportedLoginProviders'] = login_providers
        if open_id_connect_provider_arns:
            params['OpenIdConnectProviderARNs'] = open_id_connect_provider_arns
        if cognito_identity_providers:
            params['CognitoIdentityProviders'] = cognito_identity_providers
        if saml_provider_arns:
            params['SamlProviderARNs'] = saml_provider_arns

        response = self.client.create_identity_pool(**params)
        return response.get('IdentityPoolId')

    def set_role(self, identity_pool_id, authenticated_role_name,
                 unauthenticated_role_name=None):
        """ Set role for cognito identity pool authenticated and
        unauthenticated users.

        :type identity_pool_id: str
        :type authenticated_role_name: str
        :type unauthenticated_role_name: str
        """
        if not (authenticated_role_name or unauthenticated_role_name):
            return
        iam_conn = IAMConnection(None, self.aws_access_key_id,
                                 self.aws_secret_access_key,
                                 self.aws_session_token)
        params = dict(IdentityPoolId=identity_pool_id, Roles={})
        if authenticated_role_name:
            auth_role_arn = iam_conn.check_if_role_exists(
                authenticated_role_name)
            if auth_role_arn:
                params['Roles']['authenticated'] = auth_role_arn
        if unauthenticated_role_name:
            unauth_role_arn = iam_conn.check_if_role_exists(
                unauthenticated_role_name)
            if unauth_role_arn:
                params['Roles']['unauthenticated'] = unauth_role_arn
        self.client.set_identity_pool_roles(**params)

    def list_existing_pools(self):
        """ Get list of existing identity pools."""
        existing_pools = []
        identity_pools = self.client.list_identity_pools(MaxResults=60)
        existing_pools.extend(identity_pools.get('IdentityPools'))
        token = identity_pools.get('NextToken')
        while token:
            identity_pools = self.client.list_identity_pools(MaxResults=60,
                                                             NextToken=token)
            existing_pools.extend(identity_pools.get('IdentityPools'))
            token = identity_pools['NextToken']
        return existing_pools

    def if_pool_exists_by_name(self, pool_name):
        """ Check if pool exists by name.

        :type pool_name: str
        """
        pools = self.list_existing_pools()
        if pools:
            for each in pools:
                if each['IdentityPoolName'] == pool_name:
                    return each['IdentityPoolId']

    def describe_identity_pool(self, identity_pool_id):
        return self.client.describe_identity_pool(
            IdentityPoolId=identity_pool_id)

    def remove_identity_pool(self, identity_pool_id):
        """ Remove identity pool by id.

        :type identity_pool_id: str
        """
        self.client.delete_identity_pool(IdentityPoolId=identity_pool_id)

    def list_all_identities_ids_in_pool(self, identity_pool_name):
        """ Lists all identities ids in identity pool"""
        pool_id = self.if_pool_exists_by_name(identity_pool_name)
        if pool_id:
            response = self.client.list_identities(IdentityPoolId=pool_id,
                                                   MaxResults=60)
            identities_list = response['Identities']
            token = response.get('NextToken')
            while token:
                response = self.client.list_identities(IdentityPoolId=pool_id,
                                                       MaxResults=60)
                identities_list.append(response['Identities'])
                token = response['NextToken']

            identity_ids_list = []
            for identity in identities_list:
                identity_ids_list.append(identity.get('IdentityId'))
            return identity_ids_list

    def remove_specified_identities_in_pool(self, identities_to_delete):
        """ Removes specified identities from specified pool"""
        response = self.client.delete_identities(
            IdentityIdsToDelete=identities_to_delete)
        return response



import glob
import json
import re
from pathlib import Path

import click
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.constants import RESOURCES_FILE_NAME
from syndicate.core.generators import (_read_content_from_file,
                                       _write_content_to_file)

_LOG = get_logger(
    'syndicate.core.generators.deployment_resources.base_generator')
USER_LOG = get_user_logger()


class BaseConfigurationGenerator:
    """Contains params and their values for the current entity:
    - if the value of param is None, the param will be set only if it's
      given. If it isn't given, will be omitted;
    - if the value of param is set, will be used as default in case user
      doesn't put it;
    - if the value is type (dict, list, bool), the value will be the object
      of the type;
    - if the value is a dict, it will be handled and resolved resursively.
    """
    CONFIGURATION: dict = {}

    def __init__(self, **kwargs):
        self._dict = dict(kwargs)
        self.project_path = self._dict.pop('project_path')

    def _resolve_configuration(self, defaults_dict=None) -> dict:
        """Returns a dict with resolved params for current entity based on
        given params"""
        defaults_dict = defaults_dict or self.CONFIGURATION
        _LOG.info(f"Resolving not required params...")
        result = {}
        for param_name, default_value in defaults_dict.items():
            given_value = self._dict.get(param_name)
            if given_value == None or given_value == ():  # click's nulls
                if isinstance(default_value, type):
                    to_assign = default_value()
                    _LOG.info(f"Setting default value - the object "
                              f"'{to_assign}' of the class '{default_value}' "
                              f"for param '{param_name}'")
                    result[param_name] = to_assign
                elif isinstance(default_value, dict):
                    result[param_name] = \
                        self._resolve_configuration(
                            defaults_dict=default_value)
                elif default_value != None:
                    _LOG.info(f"Setting default value '{default_value}' "
                              f"for param '{param_name}'")
                    result[param_name] = default_value
            else:
                _LOG.info(f"Setting given value '{given_value}' for "
                          f"param '{param_name}'")
                if isinstance(given_value, tuple):
                    result[param_name] = list(given_value)
                else:
                    result[param_name] = given_value
        return result

    def _get_deployment_resources_files(self) -> list:
        """Returns the list of paths to each deployment_resources.json file"""
        _LOG.info(f"Recursively getting all the deployment_resources.json with"
                  f" root '{self.project_path}'")
        dep_res_files = glob.glob(str(Path(self.project_path, "**",
                                           RESOURCES_FILE_NAME)),
                                  recursive=True)
        return dep_res_files

    def _find_resources_by_type(self, resources_type) -> dict:
        """Returns the dict, where key is a path to deployment_resources file
        and value is a set of entities' names with given resource_type"""
        dep_res_files = self._get_deployment_resources_files()
        resources = {}
        _LOG.info(f"Looking for resource '{resources_type}' in meta")
        for file in dep_res_files:
            data = json.loads(_read_content_from_file(file))
            resources[file] = set(filter(
                lambda name: data[name]['resource_type'] == resources_type,
                data))
        _LOG.info(f"Found '{resources}' inside with type '{resources_type}'")
        return resources

    def _get_resource_meta_paths(self, resource_name, resource_type):
        """Returns the list of paths to deployment resouces file, where the
        resouces with given name and type is declared. (In case of api gw you
        possibly may have two declarations of the same api in different files
        """
        available_resources = self._find_resources_by_type(resource_type)
        paths = []
        _LOG.info(f"Looking for {resource_type} '{resource_name}' in meta...")
        for path, resources in available_resources.items():
            if resource_name in resources:
                _LOG.info(f"Found '{resource_name}' in meta from '{path}'")
                paths.append(path)
        _LOG.info(f"Found {resource_type} '{resource_name}' in "
                  f"these files: '{paths}'")
        return paths

    @staticmethod
    def _validate_lambda_existence(lambda_name):
        from syndicate.core import PROJECT_STATE
        _LOG.info(f"Validating existence of lambda: {lambda_name}")

        if lambda_name not in PROJECT_STATE.lambdas:
            message = f"Lambda '{lambda_name}' wasn't found"
            _LOG.error(f"Validation error: {message}")
            raise ValueError(message)
        _LOG.info(f"Validation successfully finished, lambda exists")

    def write(self):
        """The main method to write resouces"""
        raise NotImplementedError()


class BaseDeploymentResourceGenerator(BaseConfigurationGenerator):
    RESOURCE_TYPE: str = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.resource_name = self._dict.pop('resource_name').strip()
        self._validate_resource_name()

        if not self.RESOURCE_TYPE:
            message = f"RESOURCE_TYPE variable inside class " \
                      f"'{type(self).__name__}' must be specified"
            _LOG.error(message)
            raise AssertionError(message)

    def generate_deployment_resource(self) -> dict:
        """Generates resource meta for current object and returns it"""
        USER_LOG.info(f"Generating deployment resources for "
                      f"{self.RESOURCE_TYPE} '{self.resource_name}'")
        return {
            self.resource_name: self._generate_resource_configuration()
        }

    def _generate_resource_configuration(self) -> dict:
        """Return a dict with just resource configuration"""
        _LOG.info("Generating configuration for "
                  f"{self.RESOURCE_TYPE} '{self.resource_name}'")
        result = {
            "resource_type": self.RESOURCE_TYPE,
        }
        result.update(self._resolve_configuration())
        return result

    def write(self):
        """Writes generated meta to the highest, level-wise,
        deployment_resources, excluding the lambdas sub-folder. If resource
        with the name {self.resource_name} already exists, it'll ask a
        user whether overwrite it or not. If 'yes', resource meta will
        be written to the file, where the duplicate was found"""

        resources_file = next(
            iter(path for path in Path(self.project_path).rglob(
                RESOURCES_FILE_NAME) if 'lambdas' not in path.parts),
            Path(self.project_path, RESOURCES_FILE_NAME)
        )

        if not resources_file.exists():
            USER_LOG.warning('Root "deployment_resources.json" wasn\'t found. '
                             'Creating the one...')
            _write_content_to_file(resources_file, json.dumps({}))
        duplicated_file = self._find_file_with_duplicate()
        if duplicated_file:
            message = f"Resource with name '{self.resource_name}' " \
                      f"was found in file '{duplicated_file}'."
            _LOG.warning(f"Found duplicate while generating meta. {message}")
            if click.confirm(f"{message} Overwrite?"):
                USER_LOG.warning(
                    f"Overwriting resource '{self.resource_name}'")
                resources_file = duplicated_file
            else:
                USER_LOG.warning(f"Skipping resource '{self.resource_name}'")
                raise RuntimeError

        deployment_resources = json.loads(_read_content_from_file(
            resources_file
        ))
        deployment_resources.update(self.generate_deployment_resource())
        USER_LOG.info(f"Writing deployment resources for "
                      f"{self.RESOURCE_TYPE} '{self.resource_name}' "
                      f"to the file '{resources_file}'")
        _write_content_to_file(resources_file,
                               json.dumps(deployment_resources, indent=2))

    def _find_file_with_duplicate(self):
        """Looks for self.resouce_name inside each deployment_resource.json.
        If a duplicate is found, returns the path to file with it."""
        paths = self._get_resource_meta_paths(self.resource_name,
                                              self.RESOURCE_TYPE)
        if paths:
            _LOG.warning(f"Duplicated {self.RESOURCE_TYPE} with name "
                         f"'{self.resource_name}' was found in: {paths}")
            return paths[0]
        _LOG.info(f"No duplicated {self.RESOURCE_TYPE} with "
                  f"name '{self.resource_name}' was found")

    def _validate_resource_name(self):
        """Validates self.resource_name"""
        to_validate: str = self.resource_name
        _LOG.info(f"Validating resource name: '{to_validate}'")
        error = None

        invalid_character = re.search("[^a-zA-Z0-9._-]", to_validate)
        if not 3 <= len(to_validate) <= 64:
            error = "resource name length must be between 3 and 64 characters"
        elif invalid_character:
            error = f"resource name cannot contain: " \
                    f"'{invalid_character.group()}'"
        elif any(to_validate.startswith(prefix) for prefix in '_.-'):
            error = "resource name cannot start with any of these: '_', '.', '-'"
        elif any(to_validate.endswith(suffix) for suffix in '_.-'):
            error = "resource name cannot end with any of these: '_', '.', '-'"
        if error:
            _LOG.error(f"Resource name validation error: {error}")
            raise click.BadParameter(error, param_hint="resource_name")
        _LOG.info(f"Resource name: '{to_validate}' passed the validation")



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from troposphere import cognito

from .cf_resource_converter import CfResourceConverter
from ..cf_transform_utils import to_logic_name, iam_role_logic_name


class CfCognitoConverter(CfResourceConverter):

    def convert(self, name, meta):
        open_id_provider_names = meta.get('open_id_providers', [])
        account_id = self.config.account_id
        open_id_arns = ['arn:aws:iam::{0}:oidc-provider/{1}'.format(
            account_id, n) for n in open_id_provider_names]

        logic_name = to_logic_name('CognitoIdentityPool', name)
        identity_pool = cognito.IdentityPool(logic_name)
        identity_pool.AllowUnauthenticatedIdentities = False
        self.template.add_resource(identity_pool)

        provider_name = meta.get('provider_name')
        if provider_name:
            identity_pool.DeveloperProviderName = provider_name

        if open_id_arns:
            identity_pool.OpenIdConnectProviderARNs = open_id_arns

        auth_role_name = meta.get('auth_role')
        unauth_role_name = meta.get('unauth_role')
        if auth_role_name or unauth_role_name:
            role_attachment = cognito.IdentityPoolRoleAttachment(
                to_logic_name('CognitoIdentityPoolRoleAttachment', name))
            role_attachment.IdentityPoolId = identity_pool.ref()
            self.template.add_resource(role_attachment)

            roles = {}
            if auth_role_name:
                auth_role = self.get_resource(
                    iam_role_logic_name(auth_role_name))
                if auth_role:
                    roles['authenticated'] = auth_role.get_att('Arn')
            if unauth_role_name:
                unauth_role = self.get_resource(
                    iam_role_logic_name(unauth_role_name))
                if unauth_role:
                    roles['unauthenticated'] = unauth_role.get_att('Arn')
            role_attachment.Roles = roles



import copy

from syndicate.commons.log_helper import get_logger
from syndicate.core.generators.deployment_resources.base_generator import \
    BaseDeploymentResourceGenerator
from syndicate.core.constants import EC2_LAUNCH_TEMPLATE_TYPE


_LOG = get_logger('syndicate.core.generators.deployment_resources.'
                  'ec2_launch_template_generator')


class EC2LaunchTemplateGenerator(BaseDeploymentResourceGenerator):
    RESOURCE_TYPE = EC2_LAUNCH_TEMPLATE_TYPE
    CONFIGURATION = {
        'version_description': None,
        'launch_template_data': dict
    }

    KEY_MAPPING = {
        'security_group_names': 'security_groups',
        'imds_version': 'imds_support'
    }

    NON_LT_DATA_KEYS = ['version_description']

    def write(self):
        lt_data = dict()
        _dict = copy.deepcopy(self._dict)
        _LOG.debug('Going to resolve launch template data')
        for key in self._dict:
            if key in self.NON_LT_DATA_KEYS:
                continue
            if key in self.KEY_MAPPING:
                if self._dict.get(key):
                    lt_data[self.KEY_MAPPING[key]] = _dict.pop(key)
            else:
                if self._dict.get(key):
                    lt_data[key] = _dict.pop(key)
        _dict['launch_template_data'] = lt_data
        _LOG.debug(f'Launch template parameters {_dict}')
        self._dict = _dict
        super().write()
